from typing import Callable

from rx.core import Observable
from rx.core.typing import Scheduler
from rx.core.observer import ObserveOnObserver


def _observe_on(scheduler: Scheduler) -> Callable[[Observable], Observable]:
    def observe_on(source: Observable) -> Observable:
        """Wraps the source sequence in order to run its observer
        callbacks on the specified scheduler.

        This only invokes observer callbacks on a scheduler. In case
        the subscription and/or unsubscription actions have
        side-effects that require to be run on a scheduler, use
        subscribe_on.

        Args:
            source: Source observable.


        Returns:
            Returns the source sequence whose observations happen on
            the specified scheduler.
        """
        def subscribe(observer, subscribe_scheduler=None):
            return source.subscribe(ObserveOnObserver(scheduler, observer),
                                    scheduler=subscribe_scheduler)

        return Observable(subscribe)
    return observe_on



from typing import Callable, Optional

from rx import operators as ops
from rx.core import Observable, pipe, typing


def _buffer_with_time(timespan: typing.RelativeTime, timeshift: Optional[typing.RelativeTime] = None,
                      scheduler: Optional[typing.Scheduler] = None) -> Callable[[Observable], Observable]:
    if not timeshift:
        timeshift = timespan

    return pipe(
        ops.window_with_time(timespan, timeshift, scheduler),
        ops.flat_map(lambda x: x.pipe(ops.to_iterable()))
    )



from typing import Callable

import rx
from rx import Observable


def _fork_join(*args: Observable) -> Callable[[Observable], Observable]:
    def fork_join(source: Observable) -> Observable:
        """Wait for observables to complete and then combine last values
        they emitted into a tuple. Whenever any of that observables completes
        without emitting any value, result sequence will complete at that moment as well.

        Examples:
            >>> obs = fork_join(source)

        Returns:
            An observable sequence containing the result of combining last element from
            each source in given sequence.
        """
        return rx.fork_join(source, *args)

    return fork_join



from threading import Thread

from rx.core.typing import StartableTarget


def default_thread_factory(target: StartableTarget) -> Thread:
    return Thread(target=target, daemon=True)


def synchronized(lock):
    """A decorator for synchronizing access to a given function."""

    def wrapper(fn):
        def inner(*args, **kw):
            with lock:
                return fn(*args, **kw)
        return inner
    return wrapper



from typing import Callable

from rx.core import Observable
from rx.internal import noop


def _ignore_elements() -> Callable[[Observable], Observable]:
    """Ignores all elements in an observable sequence leaving only the
    termination messages.

    Returns:
        An empty observable {Observable} sequence that signals
        termination, successful or exceptional, of the source sequence.
    """

    def ignore_elements(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            return source.subscribe_(noop, observer.on_error, observer.on_completed, scheduler)

        return Observable(subscribe)
    return ignore_elements


from abc import ABC, abstractmethod


class PeriodicScheduler(ABC):
    """PeriodicScheduler abstract base class. Untyped."""

    @abstractmethod
    def schedule_periodic(self, period, action, state=None):
        return NotImplemented



from threading import Lock, Timer
from typing import MutableMapping, Optional
from weakref import WeakKeyDictionary

from rx.core import typing
from rx.disposable import CompositeDisposable, Disposable, SingleAssignmentDisposable

from .periodicscheduler import PeriodicScheduler


class TimeoutScheduler(PeriodicScheduler):
    """A scheduler that schedules work via a timed callback."""

    _lock = Lock()
    _global: MutableMapping[type, 'TimeoutScheduler'] = WeakKeyDictionary()

    @classmethod
    def singleton(cls) -> 'TimeoutScheduler':
        with TimeoutScheduler._lock:
            try:
                self = TimeoutScheduler._global[cls]
            except KeyError:
                self = super().__new__(cls)
                TimeoutScheduler._global[cls] = self
        return self

    def __new__(cls) -> 'TimeoutScheduler':
        return cls.singleton()

    def schedule(self,
                 action: typing.ScheduledAction,
                 state: Optional[typing.TState] = None
                 ) -> typing.Disposable:
        """Schedules an action to be executed.

        Args:
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        sad = SingleAssignmentDisposable()

        def interval() -> None:
            sad.disposable = self.invoke_action(action, state)

        timer = Timer(0, interval)
        timer.setDaemon(True)
        timer.start()

        def dispose() -> None:
            timer.cancel()

        return CompositeDisposable(sad, Disposable(dispose))

    def schedule_relative(self,
                          duetime: typing.RelativeTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed after duetime.

        Args:
            duetime: Relative time after which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        seconds = self.to_seconds(duetime)
        if seconds <= 0.0:
            return self.schedule(action, state)

        sad = SingleAssignmentDisposable()

        def interval() -> None:
            sad.disposable = self.invoke_action(action, state)

        timer = Timer(seconds, interval)
        timer.setDaemon(True)
        timer.start()

        def dispose() -> None:
            timer.cancel()

        return CompositeDisposable(sad, Disposable(dispose))

    def schedule_absolute(self,
                          duetime: typing.AbsoluteTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed at duetime.

        Args:
            duetime: Absolute time at which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        duetime = self.to_datetime(duetime)
        return self.schedule_relative(duetime - self.now, action, state)



from abc import ABC, abstractmethod


class Scheduler(ABC):
    """Scheduler abstract base class. Untyped."""

    @property
    @abstractmethod
    def now(self):
        return NotImplemented

    @abstractmethod
    def schedule(self, action, state=None):
        return NotImplemented

    @abstractmethod
    def schedule_relative(self, duetime, action, state=None):
        return NotImplemented

    @abstractmethod
    def schedule_absolute(self, duetime, action, state=None):
        return NotImplemented

    @classmethod
    @abstractmethod
    def to_seconds(cls, value):
        return NotImplemented

    @classmethod
    @abstractmethod
    def to_datetime(cls, value):
        return NotImplemented

    @classmethod
    @abstractmethod
    def to_timedelta(cls, value):
        return NotImplemented



from typing import Callable, Any, Optional

from rx.core import typing, Observable
from rx.core.typing import Mapper


def _to_dict(key_mapper: Mapper,
             element_mapper: Optional[Mapper] = None
             ) -> Callable[[Observable], Observable]:
    def to_dict(source: Observable) -> Observable:
        """Converts the observable sequence to a Map if it exists.

        Args:
            source: Source observable to convert.

        Returns:
            An observable sequence with a single value of a dictionary
            containing the values from the observable sequence.
        """

        def subscribe(observer: typing.Observer, scheduler: Optional[typing.Scheduler] = None) -> typing.Disposable:
            m = dict()

            def on_next(x: Any) -> None:
                try:
                    key = key_mapper(x)
                except Exception as ex:  # pylint: disable=broad-except
                    observer.on_error(ex)
                    return

                element = x
                if element_mapper:
                    try:
                        element = element_mapper(x)
                    except Exception as ex:  # pylint: disable=broad-except
                        observer.on_error(ex)
                        return

                m[key] = element

            def on_completed() -> None:
                nonlocal m
                observer.on_next(m)
                m = dict()
                observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return to_dict




from typing import Callable

from rx.core import Observable
from rx.internal import ArgumentOutOfRangeException


def _skip(count: int) -> Callable[[Observable], Observable]:
    if count < 0:
        raise ArgumentOutOfRangeException()

    def skip(source: Observable) -> Observable:
        """The skip operator.

        Bypasses a specified number of elements in an observable sequence
        and then returns the remaining elements.

        Args:
            source: The source observable.

        Returns:
            An observable sequence that contains the elements that occur
            after the specified index in the input sequence.
        """

        def subscribe(observer, scheduler=None):
            remaining = count

            def on_next(value):
                nonlocal remaining

                if remaining <= 0:
                    observer.on_next(value)
                else:
                    remaining -= 1

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return skip



import json
import os

configPath = f'{os.path.dirname(os.path.abspath(__file__))}/config.json'

class SettingsUtil:
    def __init__(self) -> None:
        self.settings : dict = dict()
        self.loadSettings()
        
    def loadSettings(self, path: str = "./config.json") -> None:
        with open(configPath, encoding='utf-8') as f:
            self.settings = json.load(f)
        return

    def setSettings(self, incomingSettings:dict, path: str = "./config.json") -> None:
        with open(configPath, "r", encoding='utf-8') as f:
            data = json.load(f)

        for key, value in incomingSettings.items():
                data["client"][key] = value

        with open(configPath, "w", encoding='utf-8') as f:
            json.dump(data, f, indent=4)
            
        self.loadSettings()
        return

    def getSettings(self) -> dict():
        return self.settings



from typing import List, Optional

import twitch.helix as helix
from twitch.api import API
from .resource import Resource


class Follows(Resource['helix.Follow']):
    FOLLOWING: int = 1
    FOLLOWED: int = 2

    def __init__(self, api: API, follow_type: 1, **kwargs: Optional):
        super().__init__(api=api, path='users/follows')
        self._kwargs = kwargs
        self.follow_type: int = follow_type

    def _can_paginate(self) -> bool:
        return True

    def _handle_pagination_response(self, response: dict) -> List['helix.Follow']:
        return [helix.Follow(api=self._api, data=follow) for follow in response.get('data', [])]

    @property
    def total(self) -> int:
        return self._api.get(self._path, params={**self._kwargs, **{'first': 100}}).get('total', -1)

    @property
    def users(self) -> 'helix.Users':
        user_ids: List[int] = []
        if self.follow_type == 'followers':
            user_ids = [int(follow.from_id) for follow in self]
        elif self.follow_type == 'followings':
            user_ids = [int(follow.to_id) for follow in self]

        return helix.Users(self._api, user_ids)



import logging

from datetime import datetime
from typing import Any, Optional

from rx.core import typing
from rx.disposable import CompositeDisposable, Disposable, SingleAssignmentDisposable

from ..periodicscheduler import PeriodicScheduler


log = logging.getLogger("Rx")


class TwistedScheduler(PeriodicScheduler):
    """A scheduler that schedules work via the Twisted reactor mainloop."""

    def __init__(self, reactor: Any) -> None:
        """Create a new TwistedScheduler.

        Args:
            reactor: The reactor to use; typically, you would get this
                by from twisted.internet import reactor
        """

        super().__init__()
        self._reactor = reactor

    def schedule(self,
                 action: typing.ScheduledAction,
                 state: Optional[typing.TState] = None
                 ) -> typing.Disposable:
        """Schedules an action to be executed.

        Args:
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return self.schedule_relative(0.0, action, state=state)

    def schedule_relative(self,
                          duetime: typing.RelativeTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed after duetime.

        Args:
            duetime: Relative time after which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        seconds = max(0.0, self.to_seconds(duetime))

        sad = SingleAssignmentDisposable()

        def interval() -> None:
            sad.disposable = action(self, state)

        log.debug("timeout: %s", seconds)
        timer = self._reactor.callLater(seconds, interval)

        def dispose() -> None:
            if not timer.called:
                timer.cancel()

        return CompositeDisposable(sad, Disposable(dispose))

    def schedule_absolute(self,
                          duetime: typing.AbsoluteTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed at duetime.

        Args:
            duetime: Absolute time at which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        duetime = self.to_datetime(duetime)
        return self.schedule_relative(duetime - self.now, action, state=state)

    @property
    def now(self) -> datetime:
        """Represents a notion of time for this scheduler. Tasks being
        scheduled on a scheduler will adhere to the time denoted by this
        property.

        Returns:
             The scheduler's current time, as a datetime instance.
        """

        return self.to_datetime(float(self._reactor.seconds()))



from typing import List, Optional

import twitch.helix as helix
from twitch.api import API
from .resource import Resource


class Games(Resource['helix.Game']):

    def __init__(self, api: API, **kwargs: Optional):
        super().__init__(api=api, path='games')

        if len(kwargs) > 0:
            self._data = [helix.Game(api=self._api, data=game) for game in
                          self._api.get(self._path, params=kwargs)['data']]

    def top(self, **kwargs) -> List['helix.Game']:
        return [helix.Game(api=self._api, data=game) for game in
                self._api.get(f'{self._path}/top', params=kwargs)['data']]

    def _can_paginate(self) -> bool:
        return False

    def _handle_pagination_response(self, response: dict) -> List['helix.Game']:
        pass



from typing import Callable

from rx import empty
from rx.core import Observable
from rx.internal import ArgumentOutOfRangeException


def _take(count: int) -> Callable[[Observable], Observable]:
    if count < 0:
        raise ArgumentOutOfRangeException()

    def take(source: Observable) -> Observable:
        """Returns a specified number of contiguous elements from the start of
        an observable sequence.

        >>> take(source)

        Keyword arguments:
        count -- The number of elements to return.

        Returns an observable sequence that contains the specified number of
        elements from the start of the input sequence.
        """

        if not count:
            return empty()

        def subscribe(observer, scheduler=None):
            remaining = count

            def on_next(value):
                nonlocal remaining

                if remaining > 0:
                    remaining -= 1
                    observer.on_next(value)
                    if not remaining:
                        observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return take



from typing import Any, Callable
from rx.core import Observable
from rx.internal.exceptions import ArgumentOutOfRangeException


def _element_at_or_default(index, has_default=False, default_value=None):
    if index < 0:
        raise ArgumentOutOfRangeException()

    def element_at_or_default(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            i = [index]

            def on_next(x):
                found = False
                with source.lock:
                    if i[0]:
                        i[0] -= 1
                    else:
                        found = True

                if found:
                    observer.on_next(x)
                    observer.on_completed()

            def on_completed():
                if not has_default:
                    observer.on_error(ArgumentOutOfRangeException())
                else:
                    observer.on_next(default_value)
                    observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return element_at_or_default



from typing import Callable, Optional, Union

import rx
from rx.core import Observable, typing
from rx.disposable import CompositeDisposable


def sample_observable(source: Observable, sampler: Observable) -> Observable:
    def subscribe(observer, scheduler=None):
        at_end = [None]
        has_value = [None]
        value = [None]

        def sample_subscribe(x=None):
            if has_value[0]:
                has_value[0] = False
                observer.on_next(value[0])

            if at_end[0]:
                observer.on_completed()

        def on_next(new_value):
            has_value[0] = True
            value[0] = new_value

        def on_completed():
            at_end[0] = True

        return CompositeDisposable(
            source.subscribe_(on_next, observer.on_error, on_completed, scheduler),
            sampler.subscribe_(sample_subscribe, observer.on_error, sample_subscribe, scheduler)
        )
    return Observable(subscribe)


def _sample(sampler: Union[typing.RelativeTime, Observable],
            scheduler: Optional[typing.Scheduler] = None
            ) -> Callable[[Observable], Observable]:

    def sample(source: Observable) -> Observable:
        """Samples the observable sequence at each interval.

        Examples:
            >>> res = sample(source)

        Args:
            source: Source sequence to sample.

        Returns:
            Sampled observable sequence.
        """

        if isinstance(sampler, typing.Observable):
            return sample_observable(source, sampler)
        else:
            return sample_observable(source, rx.interval(sampler, scheduler=scheduler))

    return sample



from typing import Callable
from rx.core import Observable
from rx.core.notification import OnNext, OnError, OnCompleted


def _materialize() -> Callable[[Observable], Observable]:
    def materialize(source: Observable) -> Observable:
        """Partially applied materialize operator.

        Materializes the implicit notifications of an observable
        sequence as explicit notification values.

        Args:
            source: Source observable to materialize.

        Returns:
            An observable sequence containing the materialized
            notification values from the source sequence.
        """

        def subscribe(observer, scheduler=None):
            def on_next(value):
                observer.on_next(OnNext(value))

            def on_error(exception):
                observer.on_next(OnError(exception))
                observer.on_completed()

            def on_completed():
                observer.on_next(OnCompleted())
                observer.on_completed()

            return source.subscribe_(on_next, on_error, on_completed, scheduler)
        return Observable(subscribe)
    return materialize



from typing import Any, Callable

from rx import defer, operators as ops
from rx.internal.utils import NotSet
from rx.core import Observable
from rx.core.typing import Accumulator

def _scan(accumulator: Accumulator, seed: Any = NotSet) -> Callable[[Observable], Observable]:
    has_seed = seed is not NotSet

    def scan(source: Observable) -> Observable:
        """Partially applied scan operator.

        Applies an accumulator function over an observable sequence and
        returns each intermediate result.

        Examples:
            >>> scanned = scan(source)

        Args:
            source: The observable source to scan.

        Returns:
            An observable sequence containing the accumulated values.
        """

        def factory(scheduler):
            has_accumulation = False
            accumulation = None

            def projection(x):
                nonlocal has_accumulation
                nonlocal accumulation

                if has_accumulation:
                    accumulation = accumulator(accumulation, x)
                else:
                    accumulation = accumulator(seed, x) if has_seed else x
                    has_accumulation = True

                return accumulation
            return source.pipe(ops.map(projection))
        return defer(factory)
    return scan



from datetime import timedelta
from typing import Optional

import twitch.tmi as tmi
from twitch.api import API


class TMI:
    BASE_URL: str = 'https://tmi.twitch.tv/'

    def __init__(self,
                 client_id: str,
                 client_secret: str = None,
                 use_cache: bool = False,
                 cache_duration: Optional[timedelta] = None,
                 handle_rate_limit: bool = True,
                 bearer_token: Optional[str] = None):
        # Format bearer token
        if bearer_token:
            bearer_token = 'Bearer ' + bearer_token.lower().lstrip('bearer').strip()

        self.api = API(TMI.BASE_URL,
                       client_id=client_id,
                       client_secret=client_secret,
                       use_cache=use_cache,
                       cache_duration=cache_duration,
                       handle_rate_limit=handle_rate_limit,
                       bearer_token=bearer_token)

    def chatters(self, user: str) -> 'tmi.Chatters':
        return tmi.Chatters(api=self.api, user=user)



import logging
import socket
import threading
from typing import List

from rx.subject import Subject


class IRC(threading.Thread):

    def __init__(self, nickname: str, password: str, address: str = 'irc.chat.twitch.tv', port: int = 6667):
        super().__init__()

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.address: str = address
        self.port: int = port
        self.channels: List[str] = []
        self.nickname: str = nickname
        self.password: str = 'oauth:' + password.lstrip('oauth:')
        self.active: bool = True
        self.incoming: Subject = Subject()

    def run(self):
        self.connect()
        self.authenticate()

        while self.active:
            try:
                data = self._read_line()
                text = data.decode("UTF-8").strip('\n\r')

                if text.find('PING') >= 0:
                    self.send_raw('PONG ' + text.split()[1])

                if text.find('Login authentication failed') > 0:
                    logging.fatal('IRC authentication error: ' + text or '')
                    return

                # Publish data to subscribers
                self.incoming.on_next(data)

            except IOError:
                break

    def send_raw(self, message: str) -> None:
        data = (message.lstrip('\n') + '\n').encode('utf-8')
        self.socket.send(data)

    def send_message(self, message: str, channel: str) -> None:
        channel = channel.lstrip('#')
        self.send_raw(f'PRIVMSG #{channel} :{message}')

    def connect(self) -> None:
        self.socket.connect((self.address, self.port))

    def authenticate(self) -> None:
        self.send_raw(f'PASS {self.password}')
        self.send_raw(f'NICK {self.nickname}')

    def join_channel(self, channel: str) -> None:
        channel = channel.lstrip('#')
        self.channels.append(channel)
        self.send_raw(f'JOIN #{channel}')

    def leave_channel(self, channel: str) -> None:
        channel = channel.lstrip('#')
        self.channels.remove(channel)
        self.send_raw(f'PART #{channel}')

    def leave_channels(self, channels: List[str]) -> None:
        channels = [channel.lstrip('#') for channel in channels]
        [self.channels.remove(channel) for channel in channels]
        self.send_raw('PART #' + '#'.join(channels))

    def _read_line(self) -> bytes:
        data: bytes = b''
        while True:
            next_byte: bytes = self.socket.recv(1)
            if next_byte == b'\n':
                break
            data += next_byte

        return data



from .disposable import Disposable
from .booleandisposable import BooleanDisposable
from .compositedisposable import CompositeDisposable
from .singleassignmentdisposable import SingleAssignmentDisposable
from .multipleassignmentdisposable import MultipleAssignmentDisposable
from .serialdisposable import SerialDisposable
from .refcountdisposable import RefCountDisposable
from .scheduleddisposable import ScheduledDisposable



from typing import Callable, Optional

import rx
from rx import operators as ops
from rx.core import Observable, typing


def _delay_subscription(duetime: typing.AbsoluteOrRelativeTime,
                        scheduler: Optional[typing.Scheduler] = None
                        ) -> Callable[[Observable], Observable]:
    def delay_subscription(source: Observable) -> Observable:
        """Time shifts the observable sequence by delaying the subscription.

        Exampeles.
            >>> res = source.delay_subscription(5)

        Args:
            source: Source subscription to delay.

        Returns:
            Time-shifted sequence.
        """

        def mapper(_) -> Observable:
            return rx.empty()

        return source.pipe(
            ops.delay_with_mapper(rx.timer(duetime, scheduler=scheduler), mapper)
        )
    return delay_subscription



from typing import Callable, Optional, Any

from rx import operators as ops
from rx.core import Observable, pipe


def _buffer(boundaries: Observable) -> Callable[[Observable], Observable]:
    return pipe(
        ops.window(boundaries),
        ops.flat_map(pipe(ops.to_iterable(), ops.map(list)))
        )


def _buffer_when(closing_mapper: Callable[[], Observable]) -> Callable[[Observable], Observable]:
    return pipe(
        ops.window_when(closing_mapper),
        ops.flat_map(pipe(ops.to_iterable(), ops.map(list)))
        )


def _buffer_toggle(openings: Observable,
                   closing_mapper: Callable[[Any], Observable]
                   ) -> Callable[[Observable], Observable]:
    return pipe(
        ops.window_toggle(openings, closing_mapper),
        ops.flat_map(pipe(ops.to_iterable(), ops.map(list)))
        )


def _buffer_with_count(count: int, skip: Optional[int] = None) -> Callable[[Observable], Observable]:
    """Projects each element of an observable sequence into zero or more
    buffers which are produced based on element count information.

    Examples:
        >>> res = buffer_with_count(10)(xs)
        >>> res = buffer_with_count(10, 1)(xs)

    Args:
        count: Length of each buffer.
        skip: [Optional] Number of elements to skip between
            creation of consecutive buffers. If not provided, defaults to
            the count.

    Returns:
        A function that takes an observable source and returns an
        observable sequence of buffers.
    """

    def buffer_with_count(source: Observable) -> Observable:
        nonlocal skip

        if skip is None:
            skip = count

        def mapper(value):
            return value.pipe(ops.to_iterable(), ops.map(list))

        def predicate(value):
            return len(value) > 0

        return source.pipe(ops.window_with_count(count, skip), ops.flat_map(mapper), ops.filter(predicate))
    return buffer_with_count



from typing import Callable, Optional
from rx.core import Observable, pipe
from rx.core.typing import Predicate

from rx import operators as ops


def _count(predicate: Optional[Predicate] = None) -> Callable[[Observable], Observable]:

    if predicate:
        filtering = ops.filter(predicate)
        return pipe(filtering, ops.count())

    counter = ops.reduce(lambda n, _: n + 1, seed=0)
    return pipe(counter)



from typing import Any, Callable

from rx import operators as ops
from rx.internal.utils import NotSet
from rx.core import Observable, pipe
from rx.core.typing import Accumulator


def _reduce(accumulator: Accumulator, seed: Any = NotSet) -> Callable[[Observable], Observable]:
    """Applies an accumulator function over an observable sequence,
    returning the result of the aggregation as a single element in the
    result sequence. The specified seed value is used as the initial
    accumulator value.

    For aggregation behavior with incremental intermediate results, see
    `scan()`.

    Examples:
        >>> res = reduce(lambda acc, x: acc + x)
        >>> res = reduce(lambda acc, x: acc + x, 0)

    Args:
        accumulator: An accumulator function to be
            invoked on each element.
        seed: Optional initial accumulator value.

    Returns:
        An operator function that takes an observable source and returns
        an observable sequence containing a single element with the
        final accumulator value.
    """
    if seed is not NotSet:
        scanner = ops.scan(accumulator, seed=seed)
        return pipe(scanner, ops.last_or_default(default_value=seed))

    return pipe(ops.scan(accumulator), ops.last())



from typing import Any
from datetime import datetime


# Defaults
def noop(*args, **kw):
    """No operation. Returns nothing"""
    pass


def identity(x: Any) -> Any:
    """Returns argument x"""
    return x


def default_now() -> datetime:
    return datetime.utcnow()


def default_comparer(x: Any, y: Any) -> bool:
    return x == y


def default_sub_comparer(x, y):
    return x - y


def default_key_serializer(x: Any) -> str:
    return str(x)


def default_error(err) -> Exception:
    if isinstance(err, BaseException):
        raise err
    else:
        raise Exception(err)



import sys
class Logger:
    @staticmethod
    def print_fail(message, end = '\n'):
        sys.stderr.write('\x1b[1;31m' + message.strip() + '\x1b[0m' + end)

    @staticmethod
    def print_pass(message, end = '\n'):
        sys.stdout.write('\x1b[1;32m' + message.strip() + '\x1b[0m' + end)

    @staticmethod
    def print_warn(message, end = '\n'):
        sys.stderr.write('\x1b[1;33m' + message.strip() + '\x1b[0m' + end)

    @staticmethod
    def print_info(message, end = '\n'):
        sys.stdout.write('\x1b[1;34m' + message.strip() + '\x1b[0m' + end)

    @staticmethod
    def print_bold(message, end = '\n'):
        sys.stdout.write('\x1b[1;37m' + message.strip() + '\x1b[0m' + end)

    @staticmethod
    def exit_line():
        sys.stdout.write("\n")


from abc import abstractmethod

from .observable import Observable
from .observer import Observer


class Subject(Observer, Observable):
    """Subject abstract base class. Untyped."""
    __slots__ = ()

    @abstractmethod
    def on_next(self, value):
        raise NotImplementedError

    @abstractmethod
    def on_error(self, error):
        raise NotImplementedError

    @abstractmethod
    def on_completed(self):
        raise NotImplementedError

    @abstractmethod
    def subscribe(self, observer=None, *, scheduler=None):
        raise NotImplementedError



from typing import Callable, Optional

import rx
from rx import from_future
from rx.core import Observable
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable
from rx.internal.concurrency import synchronized
from rx.internal.utils import is_future


def _merge(*sources: Observable,
           max_concurrent: Optional[int] = None
           ) -> Callable[[Observable], Observable]:

    def merge(source: Observable) -> Observable:
        """Merges an observable sequence of observable sequences into
        an observable sequence, limiting the number of concurrent
        subscriptions to inner sequences. Or merges two observable
        sequences into a single observable sequence.

        Examples:
            >>> res = merge(sources)

        Args:
            source: Source observable.

        Returns:
            The observable sequence that merges the elements of the
            inner sequences.
        """

        if max_concurrent is None:
            sources_ = tuple([source]) + sources
            return rx.merge(*sources_)

        def subscribe(observer, scheduler=None):
            active_count = [0]
            group = CompositeDisposable()
            is_stopped = [False]
            queue = []

            def subscribe(xs):
                subscription = SingleAssignmentDisposable()
                group.add(subscription)

                @synchronized(source.lock)
                def on_completed():
                    group.remove(subscription)
                    if queue:
                        s = queue.pop(0)
                        subscribe(s)
                    else:
                        active_count[0] -= 1
                        if is_stopped[0] and active_count[0] == 0:
                            observer.on_completed()

                on_next = synchronized(source.lock)(observer.on_next)
                on_error = synchronized(source.lock)(observer.on_error)
                subscription.disposable = xs.subscribe_(on_next, on_error, on_completed, scheduler)

            def on_next(inner_source):
                if active_count[0] < max_concurrent:
                    active_count[0] += 1
                    subscribe(inner_source)
                else:
                    queue.append(inner_source)

            def on_completed():
                is_stopped[0] = True
                if active_count[0] == 0:
                    observer.on_completed()

            group.add(source.subscribe_(on_next, observer.on_error, on_completed, scheduler))
            return group
        return Observable(subscribe)
    return merge


def _merge_all() -> Callable[[Observable], Observable]:
    def merge_all(source: Observable) -> Observable:
        """Partially applied merge_all operator.

        Merges an observable sequence of observable sequences into an
        observable sequence.

        Args:
            source: Source observable to merge.

        Returns:
            The observable sequence that merges the elements of the inner
            sequences.
        """
        def subscribe(observer, scheduler=None):
            group = CompositeDisposable()
            is_stopped = [False]
            m = SingleAssignmentDisposable()
            group.add(m)

            def on_next(inner_source):
                inner_subscription = SingleAssignmentDisposable()
                group.add(inner_subscription)

                inner_source = from_future(inner_source) if is_future(inner_source) else inner_source

                @synchronized(source.lock)
                def on_completed():
                    group.remove(inner_subscription)
                    if is_stopped[0] and len(group) == 1:
                        observer.on_completed()

                on_next = synchronized(source.lock)(observer.on_next)
                on_error = synchronized(source.lock)(observer.on_error)
                subscription = inner_source.subscribe_(on_next, on_error, on_completed, scheduler)
                inner_subscription.disposable = subscription

            def on_completed():
                is_stopped[0] = True
                if len(group) == 1:
                    observer.on_completed()

            m.disposable = source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
            return group

        return Observable(subscribe)
    return merge_all



from typing import Callable
from rx import operators as ops
from rx.core import Observable, pipe


def _is_empty() -> Callable[[Observable], Observable]:
    """Determines whether an observable sequence is empty.

    Returns:
        An observable sequence containing a single element
        determining whether the source sequence is empty.
    """

    return pipe(
        ops.some(),
        ops.map(lambda b: not b)
    )



from typing import Dict, Any

import twitch.helix as helix
from twitch.api import API
from .model import Model


class Follow(Model):

    def __init__(self, api: API, data: Dict[str, Any]):
        super().__init__(api, data)

        self.from_id: str = data.get('from_id')
        self.from_name: str = data.get('from_name')
        self.to_id: str = data.get('to_id')
        self.to_name: str = data.get('to_name')
        self.followed_at: str = data.get('followed_at')

    @property
    def follower(self) -> 'helix.User':
        """
        This user follows the followed
        :return: User following the user
        """
        return helix.Users(self._api, int(self.from_id))[0]

    @property
    def followed(self) -> 'helix.User':
        """
        This user is being followed by the follower
        :return: User being followed
        """
        return helix.Users(self._api, int(self.to_id))[0]



import os
import subprocess
import sys

import pytest

_test_timeout = 10  # Empirically, 1s is not enough on CI.

# NOTE: TkAgg tests seem to have interactions between tests,
# So isolate each test in a subprocess. See GH#18261


@pytest.mark.backend('TkAgg', skip_on_importerror=True)
def test_blit():
    script = """
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends import _tkagg
def evil_blit(photoimage, aggimage, offsets, bboxptr):
    data = np.asarray(aggimage)
    height, width = data.shape[:2]
    dataptr = (height, width, data.ctypes.data)
    _tkagg.blit(
        photoimage.tk.interpaddr(), str(photoimage), dataptr, offsets,
        bboxptr)

fig, ax = plt.subplots()
bad_boxes = ((-1, 2, 0, 2),
             (2, 0, 0, 2),
             (1, 6, 0, 2),
             (0, 2, -1, 2),
             (0, 2, 2, 0),
             (0, 2, 1, 6))
for bad_box in bad_boxes:
    try:
        evil_blit(fig.canvas._tkphoto,
                  np.ones((4, 4, 4)),
                  (0, 1, 2, 3),
                  bad_box)
    except ValueError:
        print("success")
"""
    try:
        proc = subprocess.run(
            [sys.executable, "-c", script],
            env={**os.environ,
                 "MPLBACKEND": "TkAgg",
                 "SOURCE_DATE_EPOCH": "0"},
            timeout=_test_timeout,
            stdout=subprocess.PIPE,
            check=True,
            universal_newlines=True,
        )
    except subprocess.TimeoutExpired:
        pytest.fail("Subprocess timed out")
    except subprocess.CalledProcessError:
        pytest.fail("Likely regression on out-of-bounds data access"
                    " in _tkagg.cpp")
    else:
        print(proc.stdout)
        assert proc.stdout.count("success") == 6  # len(bad_boxes)


@pytest.mark.backend('TkAgg', skip_on_importerror=True)
def test_figuremanager_preserves_host_mainloop():
    script = """
import tkinter
import matplotlib.pyplot as plt
success = False

def do_plot():
    plt.figure()
    plt.plot([1, 2], [3, 5])
    plt.close()
    root.after(0, legitimate_quit)

def legitimate_quit():
    root.quit()
    global success
    success = True

root = tkinter.Tk()
root.after(0, do_plot)
root.mainloop()

if success:
    print("success")
"""
    try:
        proc = subprocess.run(
            [sys.executable, "-c", script],
            env={**os.environ,
                 "MPLBACKEND": "TkAgg",
                 "SOURCE_DATE_EPOCH": "0"},
            timeout=_test_timeout,
            stdout=subprocess.PIPE,
            check=True,
            universal_newlines=True,
        )
    except subprocess.TimeoutExpired:
        pytest.fail("Subprocess timed out")
    except subprocess.CalledProcessError:
        pytest.fail("Subprocess failed to test intended behavior")
    else:
        assert proc.stdout.count("success") == 1


@pytest.mark.backend('TkAgg', skip_on_importerror=True)
@pytest.mark.flaky(reruns=3)
def test_figuremanager_cleans_own_mainloop():
    script = '''
import tkinter
import time
import matplotlib.pyplot as plt
import threading
from matplotlib.cbook import _get_running_interactive_framework

root = tkinter.Tk()
plt.plot([1, 2, 3], [1, 2, 5])

def target():
    while not 'tk' == _get_running_interactive_framework():
        time.sleep(.01)
    plt.close()
    if show_finished_event.wait():
        print('success')

show_finished_event = threading.Event()
thread = threading.Thread(target=target, daemon=True)
thread.start()
plt.show(block=True)  # testing if this function hangs
show_finished_event.set()
thread.join()

'''
    try:
        proc = subprocess.run(
            [sys.executable, "-c", script],
            env={**os.environ,
                 "MPLBACKEND": "TkAgg",
                 "SOURCE_DATE_EPOCH": "0"},
            timeout=_test_timeout,
            stdout=subprocess.PIPE,
            universal_newlines=True,
            check=True
        )
    except subprocess.TimeoutExpired:
        pytest.fail("Most likely plot.show(block=True) hung")
    except subprocess.CalledProcessError:
        pytest.fail("Subprocess failed to test intended behavior")
    assert proc.stdout.count("success") == 1


@pytest.mark.backend('TkAgg', skip_on_importerror=True)
@pytest.mark.flaky(reruns=3)
def test_never_update():
    script = """
import tkinter
del tkinter.Misc.update
del tkinter.Misc.update_idletasks

import matplotlib.pyplot as plt
fig = plt.figure()
plt.show(block=False)

# regression test on FigureCanvasTkAgg
plt.draw()
# regression test on NavigationToolbar2Tk
fig.canvas.toolbar.configure_subplots()

# check for update() or update_idletasks() in the event queue
# functionally equivalent to tkinter.Misc.update
# must pause >= 1 ms to process tcl idle events plus
# extra time to avoid flaky tests on slow systems
plt.pause(0.1)

# regression test on FigureCanvasTk filter_destroy callback
plt.close(fig)
"""
    try:
        proc = subprocess.run(
            [sys.executable, "-c", script],
            env={**os.environ,
                 "MPLBACKEND": "TkAgg",
                 "SOURCE_DATE_EPOCH": "0"},
            timeout=_test_timeout,
            capture_output=True,
            universal_newlines=True,
        )
    except subprocess.TimeoutExpired:
        pytest.fail("Subprocess timed out")
    else:
        # test framework doesn't see tkinter callback exceptions normally
        # see tkinter.Misc.report_callback_exception
        assert "Exception in Tkinter callback" not in proc.stderr
        # make sure we can see other issues
        print(proc.stderr, file=sys.stderr)
        # Checking return code late so the Tkinter assertion happens first
        if proc.returncode:
            pytest.fail("Subprocess failed to test intended behavior")


@pytest.mark.backend('TkAgg', skip_on_importerror=True)
def test_missing_back_button():
    script = """
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
class Toolbar(NavigationToolbar2Tk):
    # only display the buttons we need
    toolitems = [t for t in NavigationToolbar2Tk.toolitems if
                 t[0] in ('Home', 'Pan', 'Zoom')]

fig = plt.figure()
print("setup complete")
# this should not raise
Toolbar(fig.canvas, fig.canvas.manager.window)
print("success")
"""
    try:
        proc = subprocess.run(
            [sys.executable, "-c", script],
            env={**os.environ,
                 "MPLBACKEND": "TkAgg",
                 "SOURCE_DATE_EPOCH": "0"},
            timeout=_test_timeout,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        )
    except subprocess.TimeoutExpired:
        pytest.fail("Subprocess timed out")
    else:
        assert proc.stdout.count("setup complete") == 1
        assert proc.stdout.count("success") == 1
        # Checking return code late so the stdout assertions happen first
        if proc.returncode:
            pytest.fail("Subprocess failed to test intended behavior")



from typing import Callable, Any, Optional

from rx.core import Observable
from rx.core.typing import Predicate
from rx.internal.exceptions import SequenceContainsNoElementsError
from rx import operators as ops


def last_or_default_async(source: Observable,
                          has_default: bool = False,
                          default_value: Any = None
                          ) -> Observable:

    def subscribe(observer, scheduler=None):
        value = [default_value]
        seen_value = [False]

        def on_next(x):
            value[0] = x
            seen_value[0] = True

        def on_completed():
            if not seen_value[0] and not has_default:
                observer.on_error(SequenceContainsNoElementsError())
            else:
                observer.on_next(value[0])
                observer.on_completed()

        return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
    return Observable(subscribe)


def _last_or_default(predicate: Optional[Predicate] = None,
                     default_value: Any = None
                     ) -> Callable[[Observable], Observable]:

    def last_or_default(source: Observable) -> Observable:
        """Return last or default element.

        Examples:
            >>> res = _last_or_default(source)

        Args:
            source: Observable sequence to get the last item from.

        Returns:
            Observable sequence containing the last element in the
            observable sequence.
        """

        if predicate:
            return source.pipe(
                ops.filter(predicate),
                ops.last_or_default(None, default_value),
            )

        return last_or_default_async(source, True, default_value)
    return last_or_default



from collections import OrderedDict
import functools
import logging
import urllib.parse

import numpy as np

from matplotlib import _text_layout, dviread, font_manager, rcParams
from matplotlib.font_manager import FontProperties, get_font
from matplotlib.ft2font import LOAD_NO_HINTING, LOAD_TARGET_LIGHT
from matplotlib.mathtext import MathTextParser
from matplotlib.path import Path
from matplotlib.transforms import Affine2D

_log = logging.getLogger(__name__)


class TextToPath:
    """A class that converts strings to paths."""

    FONT_SCALE = 100.
    DPI = 72

    def __init__(self):
        self.mathtext_parser = MathTextParser('path')
        self._texmanager = None

    def _get_font(self, prop):
        """
        Find the `FT2Font` matching font properties *prop*, with its size set.
        """
        fname = font_manager.findfont(prop)
        font = get_font(fname)
        font.set_size(self.FONT_SCALE, self.DPI)
        return font

    def _get_hinting_flag(self):
        return LOAD_NO_HINTING

    def _get_char_id(self, font, ccode):
        """
        Return a unique id for the given font and character-code set.
        """
        return urllib.parse.quote(f"{font.postscript_name}-{ccode:x}")

    def get_text_width_height_descent(self, s, prop, ismath):
        if ismath == "TeX":
            texmanager = self.get_texmanager()
            fontsize = prop.get_size_in_points()
            w, h, d = texmanager.get_text_width_height_descent(s, fontsize,
                                                               renderer=None)
            return w, h, d

        fontsize = prop.get_size_in_points()
        scale = fontsize / self.FONT_SCALE

        if ismath:
            prop = prop.copy()
            prop.set_size(self.FONT_SCALE)
            width, height, descent, *_ = \
                self.mathtext_parser.parse(s, 72, prop)
            return width * scale, height * scale, descent * scale

        font = self._get_font(prop)
        font.set_text(s, 0.0, flags=LOAD_NO_HINTING)
        w, h = font.get_width_height()
        w /= 64.0  # convert from subpixels
        h /= 64.0
        d = font.get_descent()
        d /= 64.0
        return w * scale, h * scale, d * scale

    def get_text_path(self, prop, s, ismath=False):
        """
        Convert text *s* to path (a tuple of vertices and codes for
        matplotlib.path.Path).

        Parameters
        ----------
        prop : `~matplotlib.font_manager.FontProperties`
            The font properties for the text.

        s : str
            The text to be converted.

        ismath : {False, True, "TeX"}
            If True, use mathtext parser.  If "TeX", use tex for rendering.

        Returns
        -------
        verts : list
            A list of numpy arrays containing the x and y coordinates of the
            vertices.

        codes : list
            A list of path codes.

        Examples
        --------
        Create a list of vertices and codes from a text, and create a `.Path`
        from those::

            from matplotlib.path import Path
            from matplotlib.textpath import TextToPath
            from matplotlib.font_manager import FontProperties

            fp = FontProperties(family="Humor Sans", style="italic")
            verts, codes = TextToPath().get_text_path(fp, "ABC")
            path = Path(verts, codes, closed=False)

        Also see `TextPath` for a more direct way to create a path from a text.
        """
        if ismath == "TeX":
            glyph_info, glyph_map, rects = self.get_glyphs_tex(prop, s)
        elif not ismath:
            font = self._get_font(prop)
            glyph_info, glyph_map, rects = self.get_glyphs_with_font(font, s)
        else:
            glyph_info, glyph_map, rects = self.get_glyphs_mathtext(prop, s)

        verts, codes = [], []

        for glyph_id, xposition, yposition, scale in glyph_info:
            verts1, codes1 = glyph_map[glyph_id]
            if len(verts1):
                verts1 = np.array(verts1) * scale + [xposition, yposition]
                verts.extend(verts1)
                codes.extend(codes1)

        for verts1, codes1 in rects:
            verts.extend(verts1)
            codes.extend(codes1)

        return verts, codes

    def get_glyphs_with_font(self, font, s, glyph_map=None,
                             return_new_glyphs_only=False):
        """
        Convert string *s* to vertices and codes using the provided ttf font.
        """

        if glyph_map is None:
            glyph_map = OrderedDict()

        if return_new_glyphs_only:
            glyph_map_new = OrderedDict()
        else:
            glyph_map_new = glyph_map

        xpositions = []
        glyph_ids = []
        for item in _text_layout.layout(s, font):
            char_id = self._get_char_id(font, ord(item.char))
            glyph_ids.append(char_id)
            xpositions.append(item.x)
            if char_id not in glyph_map:
                glyph_map_new[char_id] = font.get_path()

        ypositions = [0] * len(xpositions)
        sizes = [1.] * len(xpositions)

        rects = []

        return (list(zip(glyph_ids, xpositions, ypositions, sizes)),
                glyph_map_new, rects)

    def get_glyphs_mathtext(self, prop, s, glyph_map=None,
                            return_new_glyphs_only=False):
        """
        Parse mathtext string *s* and convert it to a (vertices, codes) pair.
        """

        prop = prop.copy()
        prop.set_size(self.FONT_SCALE)

        width, height, descent, glyphs, rects = self.mathtext_parser.parse(
            s, self.DPI, prop)

        if not glyph_map:
            glyph_map = OrderedDict()

        if return_new_glyphs_only:
            glyph_map_new = OrderedDict()
        else:
            glyph_map_new = glyph_map

        xpositions = []
        ypositions = []
        glyph_ids = []
        sizes = []

        for font, fontsize, ccode, ox, oy in glyphs:
            char_id = self._get_char_id(font, ccode)
            if char_id not in glyph_map:
                font.clear()
                font.set_size(self.FONT_SCALE, self.DPI)
                font.load_char(ccode, flags=LOAD_NO_HINTING)
                glyph_map_new[char_id] = font.get_path()

            xpositions.append(ox)
            ypositions.append(oy)
            glyph_ids.append(char_id)
            size = fontsize / self.FONT_SCALE
            sizes.append(size)

        myrects = []
        for ox, oy, w, h in rects:
            vert1 = [(ox, oy), (ox, oy + h), (ox + w, oy + h),
                     (ox + w, oy), (ox, oy), (0, 0)]
            code1 = [Path.MOVETO,
                     Path.LINETO, Path.LINETO, Path.LINETO, Path.LINETO,
                     Path.CLOSEPOLY]
            myrects.append((vert1, code1))

        return (list(zip(glyph_ids, xpositions, ypositions, sizes)),
                glyph_map_new, myrects)

    def get_texmanager(self):
        """Return the cached `~.texmanager.TexManager` instance."""
        if self._texmanager is None:
            from matplotlib.texmanager import TexManager
            self._texmanager = TexManager()
        return self._texmanager

    def get_glyphs_tex(self, prop, s, glyph_map=None,
                       return_new_glyphs_only=False):
        """Convert the string *s* to vertices and codes using usetex mode."""
        # Mostly borrowed from pdf backend.

        dvifile = self.get_texmanager().make_dvi(s, self.FONT_SCALE)
        with dviread.Dvi(dvifile, self.DPI) as dvi:
            page, = dvi

        if glyph_map is None:
            glyph_map = OrderedDict()

        if return_new_glyphs_only:
            glyph_map_new = OrderedDict()
        else:
            glyph_map_new = glyph_map

        glyph_ids, xpositions, ypositions, sizes = [], [], [], []

        # Gather font information and do some setup for combining
        # characters into strings.
        for x1, y1, dvifont, glyph, width in page.text:
            font, enc = self._get_ps_font_and_encoding(dvifont.texname)
            char_id = self._get_char_id(font, glyph)

            if char_id not in glyph_map:
                font.clear()
                font.set_size(self.FONT_SCALE, self.DPI)
                # See comments in _get_ps_font_and_encoding.
                if enc is not None:
                    index = font.get_name_index(enc[glyph])
                    font.load_glyph(index, flags=LOAD_TARGET_LIGHT)
                else:
                    font.load_char(glyph, flags=LOAD_TARGET_LIGHT)
                glyph_map_new[char_id] = font.get_path()

            glyph_ids.append(char_id)
            xpositions.append(x1)
            ypositions.append(y1)
            sizes.append(dvifont.size / self.FONT_SCALE)

        myrects = []

        for ox, oy, h, w in page.boxes:
            vert1 = [(ox, oy), (ox + w, oy), (ox + w, oy + h),
                     (ox, oy + h), (ox, oy), (0, 0)]
            code1 = [Path.MOVETO,
                     Path.LINETO, Path.LINETO, Path.LINETO, Path.LINETO,
                     Path.CLOSEPOLY]
            myrects.append((vert1, code1))

        return (list(zip(glyph_ids, xpositions, ypositions, sizes)),
                glyph_map_new, myrects)

    @staticmethod
    @functools.lru_cache(50)
    def _get_ps_font_and_encoding(texname):
        tex_font_map = dviread.PsfontsMap(dviread.find_tex_file('pdftex.map'))
        psfont = tex_font_map[texname]
        if psfont.filename is None:
            raise ValueError(
                f"No usable font file found for {psfont.psname} ({texname}). "
                f"The font may lack a Type-1 version.")

        font = get_font(psfont.filename)

        if psfont.encoding:
            # If psfonts.map specifies an encoding, use it: it gives us a
            # mapping of glyph indices to Adobe glyph names; use it to convert
            # dvi indices to glyph names and use the FreeType-synthesized
            # unicode charmap to convert glyph names to glyph indices (with
            # FT_Get_Name_Index/get_name_index), and load the glyph using
            # FT_Load_Glyph/load_glyph.  (That charmap has a coverage at least
            # as good as, and possibly better than, the native charmaps.)
            enc = dviread._parse_enc(psfont.encoding)
        else:
            # If psfonts.map specifies no encoding, the indices directly
            # map to the font's "native" charmap; so don't use the
            # FreeType-synthesized charmap but the native ones (we can't
            # directly identify it but it's typically an Adobe charmap), and
            # directly load the dvi glyph indices using FT_Load_Char/load_char.
            for charmap_code in [
                    1094992451,  # ADOBE_CUSTOM.
                    1094995778,  # ADOBE_STANDARD.
            ]:
                try:
                    font.select_charmap(charmap_code)
                except (ValueError, RuntimeError):
                    pass
                else:
                    break
            else:
                _log.warning("No supported encoding in font (%s).",
                             psfont.filename)
            enc = None

        return font, enc


text_to_path = TextToPath()


class TextPath(Path):
    """
    Create a path from the text.
    """

    def __init__(self, xy, s, size=None, prop=None,
                 _interpolation_steps=1, usetex=False):
        r"""
        Create a path from the text. Note that it simply is a path,
        not an artist. You need to use the `~.PathPatch` (or other artists)
        to draw this path onto the canvas.

        Parameters
        ----------
        xy : tuple or array of two float values
            Position of the text. For no offset, use ``xy=(0, 0)``.

        s : str
            The text to convert to a path.

        size : float, optional
            Font size in points. Defaults to the size specified via the font
            properties *prop*.

        prop : `matplotlib.font_manager.FontProperties`, optional
            Font property. If not provided, will use a default
            ``FontProperties`` with parameters from the
            :ref:`rcParams <matplotlib-rcparams>`.

        _interpolation_steps : int, optional
            (Currently ignored)

        usetex : bool, default: False
            Whether to use tex rendering.

        Examples
        --------
        The following creates a path from the string "ABC" with Helvetica
        font face; and another path from the latex fraction 1/2::

            from matplotlib.textpath import TextPath
            from matplotlib.font_manager import FontProperties

            fp = FontProperties(family="Helvetica", style="italic")
            path1 = TextPath((12, 12), "ABC", size=12, prop=fp)
            path2 = TextPath((0, 0), r"$\frac{1}{2}$", size=12, usetex=True)

        Also see :doc:`/gallery/text_labels_and_annotations/demo_text_path`.
        """
        # Circular import.
        from matplotlib.text import Text

        prop = FontProperties._from_any(prop)
        if size is None:
            size = prop.get_size_in_points()

        self._xy = xy
        self.set_size(size)

        self._cached_vertices = None
        s, ismath = Text(usetex=usetex)._preprocess_math(s)
        self._vertices, self._codes = text_to_path.get_text_path(
            prop, s, ismath=ismath)
        self._should_simplify = False
        self._simplify_threshold = rcParams['path.simplify_threshold']
        self._interpolation_steps = _interpolation_steps

    def set_size(self, size):
        """Set the text size."""
        self._size = size
        self._invalid = True

    def get_size(self):
        """Get the text size."""
        return self._size

    @property
    def vertices(self):
        """
        Return the cached path after updating it if necessary.
        """
        self._revalidate_path()
        return self._cached_vertices

    @property
    def codes(self):
        """
        Return the codes
        """
        return self._codes

    def _revalidate_path(self):
        """
        Update the path if necessary.

        The path for the text is initially create with the font size of
        `~.FONT_SCALE`, and this path is rescaled to other size when necessary.
        """
        if self._invalid or self._cached_vertices is None:
            tr = (Affine2D()
                  .scale(self._size / text_to_path.FONT_SCALE)
                  .translate(*self._xy))
            self._cached_vertices = tr.transform(self._vertices)
            self._invalid = False



from typing import List, Callable

from .clip import Clip
from .follow import Follow
from .game import Game
from .model import Model
from .stream import Stream
from .user import User
from .video import Video

__all__: List[Callable] = [
    Clip,
    Follow,
    Game,
    Model,
    Stream,
    User,
    Video,
]



from typing import Any, Optional

from rx.core import typing
from rx.core import Observable

from rx.scheduler import ImmediateScheduler


def _throw(exception: Exception, scheduler: Optional[typing.Scheduler] = None) -> Observable:
    exception = exception if isinstance(exception, Exception) else Exception(exception)

    def subscribe(observer: typing.Observer, scheduler: Optional[typing.Scheduler] = None) -> typing.Disposable:
        _scheduler = scheduler or ImmediateScheduler.singleton()

        def action(scheduler: typing.Scheduler, state: Any):
            observer.on_error(exception)

        return _scheduler.schedule(action)
    return Observable(subscribe)



import threading

from rx import Observable
from rx.core.notification import OnCompleted, OnError, OnNext
from rx.scheduler import VirtualTimeScheduler
from rx.testing import TestScheduler
from rx.testing.recorded import Recorded


class TestSubscriber(object):
    def __init__(self, observable: Observable, scheduler: VirtualTimeScheduler = TestScheduler()) -> None:
        super().__init__()
        self.completed = threading.Event()
        self.messages = []

        def on_next(value) -> None:
            self.messages.append(Recorded(scheduler.clock, OnNext(value)))

        def on_error(error) -> None:
            self.messages.append(Recorded(scheduler.clock, OnError(error)))
            self.completed.set()

        def on_completed() -> None:
            self.messages.append(Recorded(scheduler.clock, OnCompleted()))
            self.completed.set()

        observable.subscribe(
            on_next=on_next,
            on_error=on_error,
            on_completed=on_completed
        )

    def results(self):
        self.completed.wait(timeout=0.1)
        return self.messages



from typing import Generator, Tuple, List

import twitch.helix as helix
from twitch.api import API
from .resource import Resource


class StreamNotFound(Exception):
    pass


class Streams(Resource['helix.Stream']):

    def __init__(self, api: API, **kwargs):
        super().__init__(api=api, path='streams')

        response: dict = self._api.get(self._path, params=kwargs)

        if response['data']:
            self._data = [helix.Stream(api=self._api, data=video) for video in
                          self._api.get(self._path, params=kwargs)['data']]
        else:
            raise StreamNotFound('No stream was found')

    @property
    def users(self) -> Generator[Tuple['helix.Stream', 'helix.User'], None, None]:
        for stream in self:
            yield stream, stream.user

    def _can_paginate(self) -> bool:
        return False

    def _handle_pagination_response(self, response: dict) -> List['helix.Stream']:
        pass



from typing import Callable

from rx.disposable import Disposable
from rx.core import ConnectableObservable, Observable


def _ref_count() -> Callable[[ConnectableObservable], Observable]:
    """Returns an observable sequence that stays connected to the
    source as long as there is at least one subscription to the
    observable sequence.
    """

    connectable_subscription = [None]
    count = [0]

    def ref_count(source: ConnectableObservable) -> Observable:
        def subscribe(observer, scheduler=None):
            count[0] += 1
            should_connect = count[0] == 1
            subscription = source.subscribe(observer, scheduler=scheduler)
            if should_connect:
                connectable_subscription[0] = source.connect(scheduler)

            def dispose():
                subscription.dispose()
                count[0] -= 1
                if not count[0]:
                    connectable_subscription[0].dispose()

            return Disposable(dispose)

        return Observable(subscribe)

    return ref_count


from typing import Callable
from rx.core import Observable


def _dematerialize() -> Callable[[Observable], Observable]:
    def dematerialize(source: Observable) -> Observable:
        """Partially applied dematerialize operator.

        Dematerializes the explicit notification values of an
        observable sequence as implicit notifications.

        Returns:
            An observable sequence exhibiting the behavior
            corresponding to the source sequence's notification values.
        """

        def subscribe(observer, scheduler=None):
            def on_next(value):
                return value.accept(observer)

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return dematerialize



import unittest
from datetime import timedelta

import twitch


class TestTwitchPython(unittest.TestCase):

    def test_cache(self):
        helix = twitch.Helix(client_id='id', bearer_token='token', use_cache=True)

        helix.api.SHARED_CACHE.set('key', {'data': 'value'})

        # Entry in shared cache
        self.assertTrue(helix.api.SHARED_CACHE.has('key'))

        # Not expired
        self.assertFalse(helix.api.SHARED_CACHE.expired('key'))

        # Expiration
        helix.api.SHARED_CACHE.set('key-with-expiration', {'data': 'value'}, duration=timedelta(seconds=-1))

        # Key is expired
        self.assertTrue(helix.api.SHARED_CACHE.expired('key-with-expiration'))

        # Unable to retrieve expired value
        self.assertFalse(helix.api.SHARED_CACHE.get('key-with-expiration'))

        # Has key, but expired
        self.assertTrue(helix.api.SHARED_CACHE.has('key-with-expiration'))

        # Clean expired keys
        helix.api.SHARED_CACHE.clean()
        self.assertFalse(helix.api.SHARED_CACHE.has('key-with-expiration'))

        # Flush cache
        helix.api.SHARED_CACHE.flush()
        self.assertFalse(helix.api.SHARED_CACHE.has('key'))



from typing import Any, Optional

from rx.internal import noop, default_error
from rx.disposable import SingleAssignmentDisposable

from .. import typing


class AutoDetachObserver(typing.Observer):

    def __init__(self,
                 on_next: Optional[typing.OnNext] = None,
                 on_error: Optional[typing.OnError] = None,
                 on_completed: Optional[typing.OnCompleted] = None
                 ) -> None:
        self._on_next = on_next or noop
        self._on_error = on_error or default_error
        self._on_completed = on_completed or noop

        self._subscription = SingleAssignmentDisposable()
        self.is_stopped = False

    def on_next(self, value: Any) -> None:
        if self.is_stopped:
            return
        self._on_next(value)

    def on_error(self, error) -> None:
        if self.is_stopped:
            return
        self.is_stopped = True

        try:
            self._on_error(error)
        finally:
            self.dispose()

    def on_completed(self) -> None:
        if self.is_stopped:
            return
        self.is_stopped = True

        try:
            self._on_completed()
        finally:
            self.dispose()

    def set_disposable(self, value: typing.Disposable):
        self._subscription.disposable = value

    subscription = property(fset=set_disposable)

    def dispose(self) -> None:
        self.is_stopped = True
        self._subscription.dispose()

    def fail(self, exn: Exception) -> bool:
        if self.is_stopped:
            return False

        self.is_stopped = True
        self._on_error(exn)
        return True



from typing import Callable, Optional
import logging

from rx.core import Observable
from rx.internal.utils import add_ref
from rx.disposable import SingleAssignmentDisposable, RefCountDisposable
from rx.internal.exceptions import ArgumentOutOfRangeException
from rx.subject import Subject

log = logging.getLogger("Rx")


def _window_with_count(count: int, skip: Optional[int] = None) -> Callable[[Observable], Observable]:
    """Projects each element of an observable sequence into zero or more
    windows which are produced based on element count information.

    Examples:
        >>> window_with_count(10)
        >>> window_with_count(10, 1)

    Args:
        count: Length of each window.
        skip: [Optional] Number of elements to skip between creation of
            consecutive windows. If not specified, defaults to the
            count.

    Returns:
        An observable sequence of windows.
    """

    if count <= 0:
        raise ArgumentOutOfRangeException()

    if skip is None:
        skip = count

    if skip <= 0:
        raise ArgumentOutOfRangeException()

    def window_with_count(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            m = SingleAssignmentDisposable()
            refCountDisposable = RefCountDisposable(m)
            n = [0]
            q = []

            def create_window():
                s = Subject()
                q.append(s)
                observer.on_next(add_ref(s, refCountDisposable))

            create_window()

            def on_next(x):
                for item in q:
                    item.on_next(x)

                c = n[0] - count + 1
                if c >= 0 and c % skip == 0:
                    s = q.pop(0)
                    s.on_completed()

                n[0] += 1
                if (n[0] % skip) == 0:
                    create_window()

            def on_error(exception):
                while q:
                    q.pop(0).on_error(exception)
                observer.on_error(exception)

            def on_completed():
                while q:
                    q.pop(0).on_completed()
                observer.on_completed()

            m.disposable = source.subscribe_(on_next, on_error, on_completed, scheduler)
            return refCountDisposable
        return Observable(subscribe)
    return window_with_count



from typing import Iterable

from rx.disposable import Disposable
from rx.core import Observable
from rx.disposable import SingleAssignmentDisposable, CompositeDisposable, SerialDisposable
from rx.scheduler import CurrentThreadScheduler


def _catch_with_iterable(sources: Iterable[Observable]) -> Observable:

    """Continues an observable sequence that is terminated by an
    exception with the next observable sequence.

    Examples:
        >>> res = catch([xs, ys, zs])
        >>> res = rx.catch(src for src in [xs, ys, zs])

    Args:
        sources: an Iterable of observables. Thus a generator is accepted.

    Returns:
        An observable sequence containing elements from consecutive
        source sequences until a source sequence terminates
        successfully.
    """

    sources_ = iter(sources)

    def subscribe(observer, scheduler_=None):
        _scheduler = scheduler_ or CurrentThreadScheduler.singleton()

        subscription = SerialDisposable()
        cancelable = SerialDisposable()
        last_exception = None
        is_disposed = False

        def action(action1, state=None):
            def on_error(exn):
                nonlocal last_exception
                last_exception = exn
                cancelable.disposable = _scheduler.schedule(action)

            if is_disposed:
                return

            try:
                current = next(sources_)
            except StopIteration:
                if last_exception:
                    observer.on_error(last_exception)
                else:
                    observer.on_completed()
            except Exception as ex:  # pylint: disable=broad-except
                observer.on_error(ex)
            else:
                d = SingleAssignmentDisposable()
                subscription.disposable = d
                d.disposable = current.subscribe_(observer.on_next, on_error, observer.on_completed, scheduler_)

        cancelable.disposable = _scheduler.schedule(action)

        def dispose():
            nonlocal is_disposed
            is_disposed = True
        return CompositeDisposable(subscription, cancelable, Disposable(dispose))
    return Observable(subscribe)



# Rx Exceptions


class SequenceContainsNoElementsError(Exception):
    def __init__(self, msg=None):
        super(SequenceContainsNoElementsError, self).__init__(msg or "Sequence contains no elements")


class ArgumentOutOfRangeException(ValueError):
    def __init__(self, msg=None):
        super(ArgumentOutOfRangeException, self).__init__(msg or "Argument out of range")


class DisposedException(Exception):
    def __init__(self, msg=None):
        super(DisposedException, self).__init__(msg or "Object has been disposed")


class ReEntracyException(Exception):
    def __init__(self, msg=None):
        super(ReEntracyException, self).__init__(msg or 'Re-entrancy detected')


class CompletedException(Exception):
    def __init__(self, msg=None):
        super(CompletedException, self).__init__(msg or 'Observer completed')


class WouldBlockException(Exception):
    def __init__(self, msg=None):
        super(WouldBlockException, self).__init__(msg or "Would block")



from typing import Any
from rx.internal.basic import default_comparer


class Recorded:
    def __init__(self, time: int, value: Any, comparer=None):
        self.time = time
        self.value = value
        self.comparer = comparer or default_comparer

    def __eq__(self, other):
        """Returns true if a recorded value matches another recorded value"""

        time_match = self.time == other.time
        return time_match and self.comparer(self.value, other.value)

    equals = __eq__

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "%s@%s" % (self.value, self.time)



from abc import ABC, abstractmethod


class Observer(ABC):
    """Observer abstract base class. Untyped."""

    __slots__ = ()

    @abstractmethod
    def on_next(self, value):
        raise NotImplementedError

    @abstractmethod
    def on_error(self, error):
        raise NotImplementedError

    @abstractmethod
    def on_completed(self):
        raise NotImplementedError



from typing import Any, Callable

from rx import defer, from_future, of
from rx.core import Observable
from rx.core.typing import Accumulator
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable
from rx.internal.concurrency import synchronized
from rx.internal.utils import NotSet, is_future


def _merge_scan(accumulator: Accumulator, seed: Any = NotSet) -> Callable[[Observable], Observable]:

    def merge_scan(source: Observable) -> Observable:
        """Partially applied merge_scan operator.

        Applies an accumulator function, which returns an observable sequence,
        over an observable sequence and returns each intermediate result.

        Examples:
            >>> scanned = merge_scan(source)

        Args:
            source: The observable source to scan.

        Returns:
            An observable sequence containing the accumulated values.
        """
        def subscribe(observer, scheduler=None):
            accumulator_value = [seed]
            active = [False]
            group = CompositeDisposable()
            is_stopped = [False]
            queue = []

            def subscribe(xs):
                subscription = SingleAssignmentDisposable()
                group.add(subscription)

                @synchronized(source.lock)
                def on_next(next_accumulator_value):
                    accumulator_value[0] = next_accumulator_value
                    observer.on_next(next_accumulator_value)

                @synchronized(source.lock)
                def on_completed():
                    group.remove(subscription)
                    if queue:
                        s = queue.pop(0)
                        subscribe(s)
                    else:
                        active[0] = False
                        if is_stopped[0]:
                            observer.on_completed()

                on_error = synchronized(source.lock)(observer.on_error)
                subscription.disposable = xs.subscribe_(on_next, on_error, on_completed, scheduler)

            def on_next(value):
                def accumulate():
                    has_accumulator_value = accumulator_value[0] is not NotSet
                    if has_accumulator_value:
                        acc_source = accumulator(accumulator_value[0], value)
                        return from_future(acc_source) if is_future(acc_source) else acc_source
                    else:
                        return of(value)

                accumulator_source = defer(lambda _: accumulate())
                if not active[0]:
                    active[0] = True
                    subscribe(accumulator_source)
                else:
                    queue.append(accumulator_source)

            def on_completed():
                is_stopped[0] = True
                if not active[0]:
                    observer.on_completed()

            group.add(source.subscribe_(on_next, observer.on_error, on_completed, scheduler))
            return group
        return Observable(subscribe)
    return merge_scan




from typing import Callable, Optional

import rx
from rx import operators as ops
from rx.core import Observable
from rx.core.typing import Mapper
from rx.subject import Subject


def _group_by(key_mapper: Mapper,
              element_mapper: Optional[Mapper] = None,
              subject_mapper: Optional[Callable[[], Subject]] = None,
              ) -> Callable[[Observable], Observable]:

    def duration_mapper(_):
        return rx.never()

    return ops.group_by_until(key_mapper, element_mapper, duration_mapper, subject_mapper)



import logging

from datetime import timedelta
from typing import Any, Optional, Set

from rx.core import typing
from rx.disposable import CompositeDisposable, Disposable, SingleAssignmentDisposable

from ..periodicscheduler import PeriodicScheduler


log = logging.getLogger(__name__)


class QtScheduler(PeriodicScheduler):
    """A scheduler for a PyQt5/PySide2 event loop."""

    def __init__(self, qtcore: Any):
        """Create a new QtScheduler.

        Args:
            qtcore: The QtCore instance to use; typically you would get this by
                either import PyQt5.QtCore or import PySide2.QtCore
        """
        super().__init__()
        self._qtcore = qtcore
        timer_class: Any = self._qtcore.QTimer
        self._periodic_timers: Set[timer_class] = set()

    def schedule(self,
                 action: typing.ScheduledAction,
                 state: Optional[typing.TState] = None
                 ) -> typing.Disposable:
        """Schedules an action to be executed.

        Args:
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """
        return self.schedule_relative(0.0, action, state=state)

    def schedule_relative(self,
                          duetime: typing.RelativeTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed after duetime.

        Args:
            duetime: Relative time after which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """
        msecs = max(0, int(self.to_seconds(duetime) * 1000.0))
        sad = SingleAssignmentDisposable()
        is_disposed = False

        def invoke_action() -> None:
            if not is_disposed:
                sad.disposable = action(self, state)

        log.debug("relative timeout: %sms", msecs)

        # Use static method, let Qt C++ handle QTimer lifetime
        self._qtcore.QTimer.singleShot(msecs, invoke_action)

        def dispose() -> None:
            nonlocal is_disposed
            is_disposed = True

        return CompositeDisposable(sad, Disposable(dispose))

    def schedule_absolute(self,
                          duetime: typing.AbsoluteTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed at duetime.

        Args:
            duetime: Absolute time at which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        delta: timedelta = self.to_datetime(duetime) - self.now
        return self.schedule_relative(delta, action, state=state)

    def schedule_periodic(self,
                          period: typing.RelativeTime,
                          action: typing.ScheduledPeriodicAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules a periodic piece of work to be executed in the loop.

       Args:
            period: Period in seconds for running the work repeatedly.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """
        msecs = max(0, int(self.to_seconds(period) * 1000.0))
        sad = SingleAssignmentDisposable()

        def interval() -> None:
            nonlocal state
            state = action(state)

        log.debug("periodic timeout: %sms", msecs)

        timer = self._qtcore.QTimer()
        timer.setSingleShot(not period)
        timer.timeout.connect(interval)
        timer.setInterval(msecs)
        self._periodic_timers.add(timer)
        timer.start()

        def dispose() -> None:
            timer.stop()
            self._periodic_timers.remove(timer)
            timer.deleteLater()

        return CompositeDisposable(sad, Disposable(dispose))



from typing import List, Optional

from rx.core import Observable
from rx.core.typing import Scheduler, RelativeTime
from rx.scheduler import NewThreadScheduler

new_thread_scheduler = NewThreadScheduler()


def _to_marbles(scheduler: Optional[Scheduler] = None, timespan: RelativeTime = 0.1):

    def to_marbles(source: Observable) -> Observable:
        """Convert an observable sequence into a marble diagram string.

        Args:
            timespan: [Optional] duration of each character in second.
                If not specified, defaults to 0.1s.
            scheduler: [Optional] The scheduler used to run the the input
                sequence on.

        Returns:
            Observable stream.
        """

        def subscribe(observer, scheduler=None):
            scheduler = scheduler or new_thread_scheduler

            result: List[str] = []
            last = scheduler.now

            def add_timespan():
                nonlocal last

                now = scheduler.now
                diff = now - last
                last = now
                secs = scheduler.to_seconds(diff)
                dashes = "-" * int((secs + timespan / 2.0) * (1.0 / timespan))
                result.append(dashes)

            def on_next(value):
                add_timespan()
                result.append(stringify(value))

            def on_error(exception):
                add_timespan()
                result.append(stringify(exception))
                observer.on_next("".join(n for n in result))
                observer.on_completed()

            def on_completed():
                add_timespan()
                result.append("|")
                observer.on_next("".join(n for n in result))
                observer.on_completed()

            return source.subscribe_(on_next, on_error, on_completed)
        return Observable(subscribe)
    return to_marbles


def stringify(value):
    """Utility for stringifying an event.
    """
    string = str(value)
    if len(string) > 1:
        string = "(%s)" % string

    return string



from .priorityqueue import PriorityQueue
from .basic import noop, default_error, default_comparer
from .exceptions import SequenceContainsNoElementsError, ArgumentOutOfRangeException, DisposedException
from . import concurrency
from . import constants



from typing import Callable, Any, Optional

from rx.internal.basic import identity
from rx.internal.utils import infinite

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Mapper, MapperIndexed, Observer, Disposable, Scheduler


# pylint: disable=redefined-builtin
def _map(mapper: Optional[Mapper] = None) -> Callable[[Observable], Observable]:

    _mapper = mapper or identity

    def map(source: Observable) -> Observable:
        """Partially applied map operator.

        Project each element of an observable sequence into a new form
        by incorporating the element's index.

        Example:
            >>> map(source)

        Args:
            source: The observable source to transform.

        Returns:
            Returns an observable sequence whose elements are the
            result of invoking the transform function on each element
            of the source.
        """

        def subscribe(obv: Observer, scheduler: Scheduler = None) -> Disposable:
            def on_next(value: Any) -> None:
                try:
                    result = _mapper(value)
                except Exception as err:  # pylint: disable=broad-except
                    obv.on_error(err)
                else:
                    obv.on_next(result)

            return source.subscribe_(on_next, obv.on_error, obv.on_completed, scheduler)
        return Observable(subscribe)
    return map


def _map_indexed(mapper_indexed: Optional[MapperIndexed] = None) -> Callable[[Observable], Observable]:

    def _identity(value: Any, _: int) -> Any:
        return value

    _mapper_indexed = mapper_indexed or _identity

    return pipe(
        ops.zip_with_iterable(infinite()),
        ops.starmap_indexed(_mapper_indexed)
    )



from typing import List, Callable

from twitch.chat import Chat
from twitch.helix import Helix
from twitch.tmi import TMI
from twitch.v5 import V5

name: str = "twitch"

__all__: List[Callable] = [
    Helix,
    V5,
    TMI,
    Chat,
]



from typing import Callable

import rx
from rx.core import Observable
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable
from rx.internal.utils import is_future


def _exclusive() -> Callable[[Observable], Observable]:
    """Performs a exclusive waiting for the first to finish before
    subscribing to another observable. Observables that come in between
    subscriptions will be dropped on the floor.

    Returns:
        An exclusive observable with only the results that
        happen when subscribed.
    """

    def exclusive(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            has_current = [False]
            is_stopped = [False]
            m = SingleAssignmentDisposable()
            g = CompositeDisposable()

            g.add(m)

            def on_next(inner_source):
                if not has_current[0]:
                    has_current[0] = True

                    inner_source = rx.from_future(inner_source) if is_future(inner_source) else inner_source

                    inner_subscription = SingleAssignmentDisposable()
                    g.add(inner_subscription)

                    def on_completed_inner():
                        g.remove(inner_subscription)
                        has_current[0] = False
                        if is_stopped[0] and len(g) == 1:
                            observer.on_completed()

                    inner_subscription.disposable = inner_source.subscribe_(
                        observer.on_next,
                        observer.on_error,
                        on_completed_inner,
                        scheduler
                    )

            def on_completed():
                is_stopped[0] = True
                if not has_current[0] and len(g) == 1:
                    observer.on_completed()

            m.disposable = source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
            return g
        return Observable(subscribe)
    return exclusive



from rx.core import Observable
import math


def determine_median(sorted_list):
    if len(sorted_list) == 0:
        raise Exception("The input sequence was empty")

    if len(sorted_list) % 2 == 1:
        return sorted_list[int((len(sorted_list) + 1) / 2) - 1]
    else:
        median_1 = sorted_list[int((len(sorted_list) + 1) / 2) - 1]
        median_2 = sorted_list[int((len(sorted_list) + 1) / 2)]
        return float(median_1 + median_2) / 2.0


def median(source: Observable) -> Observable:
    """
    Calculates the statistical median on numerical emissions. The sequence must be finite.
    """
    return source.to_sorted_list().map(lambda l: determine_median(l))


def mode(source: Observable) -> Observable:
    """
    Returns the most frequently emitted value (or "values" if they have the same number of occurrences).
    The sequence must be finite.
    """
    return source.group_by(lambda v: v) \
        .flat_map(lambda grp: grp.count().map(lambda ct: (grp.key, ct))) \
        .to_sorted_list(lambda t: t[1], reverse=True) \
        .flat_map(lambda l: Observable.from_(l).take_while(lambda t: t[1] == l[0][1])) \
        .map(lambda t: t[0])


def variance(source: Observable) -> Observable:
    """
    Returns the statistical variance of the numerical emissions.
    The sequence must be finite.
    """
    squared_values = source.to_list() \
        .flat_map(lambda l: Observable.from_(l).average().flat_map(lambda avg: Observable.from_(l).map(lambda i: i - avg))) \
        .map(lambda i: i * i) \
        .publish() \
        .auto_connect(2)

    return Observable.zip(squared_values.sum(), squared_values.count(), lambda sum, ct: sum / (ct - 1))


def standard_deviation(source: Observable) -> Observable:
    """
    Returns the standard deviation of the numerical emissions:
    The sequence must be finite.
    """
    return source.variance().map(lambda i: math.sqrt(i))





from typing import Optional, Union, Mapping, Callable, Any
from asyncio import Future

from rx import empty, defer, from_future
from rx.core import Observable
from rx.internal.utils import is_future


def _case(mapper: Callable[[], Any],
          sources: Mapping,
          default_source: Optional[Union[Observable, Future]] = None
          ) -> Observable:

    default_source = default_source or empty()

    def factory(_) -> Observable:
        try:
            result = sources[mapper()]
        except KeyError:
            result = default_source

        result = from_future(result) if is_future(result) else result

        return result
    return defer(factory)



from typing import Optional

from rx.core import Observable, typing
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable


def _fork_join(*sources: Observable) -> Observable:
    """Wait for observables to complete and then combine last values
    they emitted into a tuple. Whenever any of that observables completes
    without emitting any value, result sequence will complete at that moment as well.

    Examples:
        >>> obs = rx.fork_join(obs1, obs2, obs3)

    Returns:
        An observable sequence containing the result of combining last element from
        each source in given sequence.
    """

    parent = sources[0]

    def subscribe(observer: typing.Observer,
                  scheduler: Optional[typing.Scheduler] = None
                  ) -> CompositeDisposable:
        n = len(sources)
        values = [None] * n
        is_done = [False] * n
        has_value = [False] * n

        def done(i: int):
            is_done[i] = True

            if not has_value[i]:
                observer.on_completed()
                return

            if all(is_done):
                if all(has_value):
                    observer.on_next(tuple(values))
                    observer.on_completed()
                else:
                    observer.on_completed()

        subscriptions = [None] * n

        def _subscribe(i: int):
            subscriptions[i] = SingleAssignmentDisposable()

            def on_next(value):
                with parent.lock:
                    values[i] = value
                    has_value[i] = True

            def on_completed():
                with parent.lock:
                    done(i)

            subscriptions[i].disposable = sources[i].subscribe_(
                on_next,
                observer.on_error,
                on_completed,
                scheduler
            )

        for i in range(n):
            _subscribe(i)
        return CompositeDisposable(subscriptions)

    return Observable(subscribe)



from asyncio import Future
from datetime import datetime
from typing import cast, Callable, Optional, Union

from rx import from_future, throw
from rx.core import Observable, typing
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable, SerialDisposable
from rx.scheduler import TimeoutScheduler
from rx.internal.utils import is_future


def _timeout(duetime: typing.AbsoluteTime,
             other: Optional[Union[Observable, Future]] = None,
             scheduler: Optional[typing.Scheduler] = None
             ) -> Callable[[Observable], Observable]:

    other = other or throw(Exception("Timeout"))
    if is_future(other):
        obs = from_future(cast(Future, other))
    else:
        obs = cast(Observable, other)

    def timeout(source: Observable) -> Observable:
        """Returns the source observable sequence or the other observable
        sequence if duetime elapses.

        Examples:
            >>> res = timeout(source)

        Args:
            source: Source observable to timeout

        Returns:
            An obserable sequence switching to the other sequence in
            case of a timeout.
        """
        def subscribe(observer, scheduler_=None):
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()

            if isinstance(duetime, datetime):
                scheduler_method = _scheduler.schedule_absolute
            else:
                scheduler_method = _scheduler.schedule_relative

            switched = [False]
            _id = [0]

            original = SingleAssignmentDisposable()
            subscription = SerialDisposable()
            timer = SerialDisposable()
            subscription.disposable = original

            def create_timer():
                my_id = _id[0]

                def action(scheduler, state=None):
                    switched[0] = (_id[0] == my_id)
                    timer_wins = switched[0]
                    if timer_wins:
                        subscription.disposable = obs.subscribe(observer, scheduler=scheduler)

                timer.disposable = scheduler_method(duetime, action)

            create_timer()

            def on_next(value):
                send_wins = not switched[0]
                if send_wins:
                    _id[0] += 1
                    observer.on_next(value)
                    create_timer()

            def on_error(error):
                on_error_wins = not switched[0]
                if on_error_wins:
                    _id[0] += 1
                    observer.on_error(error)

            def on_completed():
                on_completed_wins = not switched[0]
                if on_completed_wins:
                    _id[0] += 1
                    observer.on_completed()

            original.disposable = source.subscribe_(on_next, on_error, on_completed, scheduler_)
            return CompositeDisposable(subscription, timer)
        return Observable(subscribe)
    return timeout



import time
from datetime import timedelta
from typing import Dict, Any, Optional

import requests

from .cache import Cache


class API:
    SHARED_CACHE: Cache = Cache()

    def __init__(self,
                 base_url: Optional[str] = None,
                 client_id: Optional[str] = None,
                 client_secret: Optional[str] = None,
                 use_cache: Optional[bool] = False,
                 request_rate: Optional[int] = None,
                 bearer_token: Optional[str] = None,
                 handle_rate_limit: bool = True,
                 cache_duration: Optional[timedelta] = None):
        """
        Twitch API
        :param base_url: API URL
        :param client_id: Twitch Client ID
        :param use_cache: Use local API cache
        :param bearer_token: Twitch bearer token
        :param handle_rate_limit: Handle rate limits by sleeping
        :param cache_duration: Local cache duration
        """
        self.base_url: Optional[str] = base_url
        self.client_id: Optional[str] = client_id
        self.client_secret: Optional[str] = client_secret
        self.use_cache: bool = use_cache
        self.request_rate: Optional[int] = request_rate
        self.bearer_token: Optional[str] = bearer_token
        self.handle_rate_limit: bool = handle_rate_limit
        self.cache_duration: Optional[timedelta] = cache_duration

        # Rate limit
        self.rate_limit_points: int = 800 if self.bearer_token else 30
        self.rate_limit_remaining: int = self.rate_limit_points
        self.rate_limit_reset: int = 0

    def _headers(self, custom: Dict[str, str] = None) -> Dict[str, str]:
        default: Dict[str, str] = {}

        if self.client_id:
            default['Client-ID'] = self.client_id

        if self.bearer_token:
            default['Authorization'] = self.bearer_token

        return {**default, **custom} if custom else default.copy()

    def _url(self, path: str = '') -> str:
        return self.base_url.rstrip('/') + '/' + path.lstrip('/')

    @staticmethod
    def flush_cache():
        API.SHARED_CACHE.flush()

    def _handle_rate_limit(self) -> None:
        if self.handle_rate_limit and self.rate_limit_remaining == 0:
            time_to_sleep: float = min((self.rate_limit_reset - time.time()), 10)
            time_to_sleep = max(time_to_sleep, 1)

            time.sleep(time_to_sleep)

    def _set_rate_limit(self, response: requests.Response) -> None:
        # Update rate limit fields
        if 'Ratelimit-Limit' in response.headers.keys():
            self.rate_limit_points: int = int(response.headers.get('Ratelimit-Limit'))
            self.rate_limit_remaining: int = int(response.headers.get('Ratelimit-Remaining'))
            self.rate_limit_reset: int = int(response.headers.get('Ratelimit-Reset'))

    def request(self, method, path: str = '', ignore_cache: bool = False, **kwargs) -> dict:
        url: str = self._url(path=path)
        request = requests.Request(method, url, **kwargs).prepare()
        cache_key: str = f'{method}:{request.url}'

        # Cache lookup
        if self.use_cache and not ignore_cache and API.SHARED_CACHE.get(cache_key):
            return API.SHARED_CACHE.get(cache_key)

        # Check rate limit
        self._handle_rate_limit()

        while True:
            response = requests.Session().send(request)
            self._set_rate_limit(response)

            # Too many requests status
            if response.status_code == 429 and self.handle_rate_limit:
                self._handle_rate_limit()
            else:
                break

        # Raise exception if status code is not 200
        response.raise_for_status()

        # Cache response
        if self.use_cache and not ignore_cache:
            API.SHARED_CACHE.set(key=cache_key, value=response.json(), duration=self.cache_duration)

        return response.json()

    def get(self, path: str, params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, Any]] = None,
            ignore_cache: bool = False,
            **kwargs) -> dict:
        return self.request('GET', path, ignore_cache, params=params, headers=self._headers(headers), **kwargs)

    def post(self):
        pass

    def put(self):
        pass



import logging

from typing import cast, Any, Optional, Set

from rx.core import typing
from rx.disposable import CompositeDisposable, Disposable, SingleAssignmentDisposable

from ..periodicscheduler import PeriodicScheduler


log = logging.getLogger("Rx")


class WxScheduler(PeriodicScheduler):
    """A scheduler for a wxPython event loop."""

    def __init__(self, wx: Any) -> None:
        """Create a new WxScheduler.

        Args:
            wx: The wx module to use; typically, you would get this by
                import wx
        """

        super().__init__()
        self._wx = wx
        timer_class: Any = self._wx.Timer

        class Timer(timer_class):

            def __init__(self, callback) -> None:
                super().__init__()
                self.callback = callback

            def Notify(self):
                self.callback()

        self._timer_class = Timer
        self._timers: Set[Timer] = set()

    def cancel_all(self) -> None:
        """Cancel all scheduled actions.

        Should be called when destroying wx controls to prevent
        accessing dead wx objects in actions that might be pending.
        """
        for timer in self._timers:
            timer.Stop()

    def _wxtimer_schedule(self,
                          time: typing.AbsoluteOrRelativeTime,
                          action: typing.ScheduledSingleOrPeriodicAction,
                          state: Optional[typing.TState] = None,
                          periodic: bool = False
                          ) -> typing.Disposable:
        scheduler = self

        sad = SingleAssignmentDisposable()

        def interval() -> None:
            nonlocal state
            if periodic:
                state = cast(typing.ScheduledPeriodicAction, action)(state)
            else:
                sad.disposable = cast(typing.ScheduledAction, action)(scheduler, state)

        msecs = max(1, int(self.to_seconds(time) * 1000.0))  # Must be non-zero

        log.debug("timeout wx: %s", msecs)

        timer = self._timer_class(interval)
        timer.Start(
            msecs,
            self._wx.TIMER_CONTINUOUS if periodic else self._wx.TIMER_ONE_SHOT
        )
        self._timers.add(timer)

        def dispose() -> None:
            timer.Stop()
            self._timers.remove(timer)

        return CompositeDisposable(sad, Disposable(dispose))

    def schedule(self,
                 action: typing.ScheduledAction,
                 state: Optional[typing.TState] = None
                 ) -> typing.Disposable:
        """Schedules an action to be executed.

        Args:
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return self._wxtimer_schedule(0.0, action, state=state)

    def schedule_relative(self,
                          duetime: typing.RelativeTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed after duetime.

        Args:
            duetime: Relative time after which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """
        return self._wxtimer_schedule(duetime, action, state=state)

    def schedule_absolute(self,
                          duetime: typing.AbsoluteTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed at duetime.

        Args:
            duetime: Absolute time at which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        duetime = self.to_datetime(duetime)
        return self._wxtimer_schedule(duetime - self.now, action, state=state)

    def schedule_periodic(self,
                          period: typing.RelativeTime,
                          action: typing.ScheduledPeriodicAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules a periodic piece of work to be executed in the loop.

       Args:
            period: Period in seconds for running the work repeatedly.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return self._wxtimer_schedule(period, action, state=state, periodic=True)



from asyncio import Future
from typing import cast, Callable, Union

from rx import from_future
from rx.core import Observable
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable
from rx.internal.utils import is_future


def _skip_until(other: Union[Observable, Future]) -> Callable[[Observable], Observable]:
    """Returns the values from the source observable sequence only after
    the other observable sequence produces a value.

    Args:
        other: The observable sequence that triggers propagation of
            elements of the source sequence.

    Returns:
        An observable sequence containing the elements of the source
    sequence starting from the point the other sequence triggered
    propagation.
    """

    if is_future(other):
        obs = from_future(cast(Future, other))
    else:
        obs = cast(Observable, other)

    def skip_until(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            is_open = [False]

            def on_next(left):
                if is_open[0]:
                    observer.on_next(left)

            def on_completed():
                if is_open[0]:
                    observer.on_completed()

            subs = source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
            subscriptions = CompositeDisposable(subs)

            right_subscription = SingleAssignmentDisposable()
            subscriptions.add(right_subscription)

            def on_next2(x):
                is_open[0] = True
                right_subscription.dispose()

            def on_completed2():
                right_subscription.dispose()

            right_subscription.disposable = obs.subscribe_(on_next2, observer.on_error, on_completed2, scheduler)

            return subscriptions
        return Observable(subscribe)
    return skip_until



from abc import ABC, abstractmethod


class Disposable(ABC):
    """Disposable abstract base class. Untyped."""

    @abstractmethod
    def dispose(self):
        raise NotImplementedError

    def __enter__(self):
        """Context management protocol."""

    def __exit__(self, typ, value, traceback):
        """Context management protocol."""
        self.dispose()



from typing import Callable

import rx
from rx.core import Observable
from rx.core import typing
from rx.disposable import CompositeDisposable, Disposable


def _using(resource_factory: Callable[[], typing.Disposable],
           observable_factory: Callable[[typing.Disposable], Observable]
           ) -> Observable:
    """Constructs an observable sequence that depends on a resource
    object, whose lifetime is tied to the resulting observable
    sequence's lifetime.

    Example:
        >>> res = rx.using(lambda: AsyncSubject(), lambda: s: s)

    Args:
        resource_factory: Factory function to obtain a resource object.
        observable_factory: Factory function to obtain an observable
            sequence that depends on the obtained resource.

    Returns:
        An observable sequence whose lifetime controls the lifetime
        of the dependent resource object.
    """

    def subscribe(observer, scheduler=None):
        disp = Disposable()

        try:
            resource = resource_factory()
            if resource is not None:
                disp = resource

            source = observable_factory(resource)
        except Exception as exception:  # pylint: disable=broad-except
            d = rx.throw(exception).subscribe(observer, scheduler=scheduler)
            return CompositeDisposable(d, disp)

        return CompositeDisposable(source.subscribe(observer, scheduler=scheduler), disp)
    return Observable(subscribe)



from typing import List, Callable

from twitch.v5.models.comment import Comment
from twitch.v5.resources.comments import Comments
from twitch.v5.v5 import V5

__all__: List[Callable] = [
    V5,
    Comment, Comments
]



from typing import Callable, NamedTuple, Any, Optional
from datetime import datetime

from rx import defer
from rx.core import Observable, typing
from rx.scheduler import TimeoutScheduler
from rx import operators


class Timestamp(NamedTuple):
    value: Any
    timestamp: datetime


def _timestamp(scheduler: Optional[typing.Scheduler] = None) -> Callable[[Observable], Observable]:
    def timestamp(source: Observable) -> Observable:
        """Records the timestamp for each value in an observable sequence.

        Examples:
            >>> timestamp(source)

        Produces objects with attributes `value` and `timestamp`, where
        value is the original value.

        Args:
            source: Observable source to timestamp.

        Returns:
            An observable sequence with timestamp information on values.
        """

        def factory(scheduler_=None):
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()
            mapper = operators.map(lambda value: Timestamp(value=value, timestamp=_scheduler.now))

            return source.pipe(mapper)
        return defer(factory)
    return timestamp



from typing import Callable

from rx.core import Observable, typing
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable, SerialDisposable


def _delay_with_mapper(subscription_delay=None, delay_duration_mapper=None) -> Callable[[Observable], Observable]:
    def delay_with_mapper(source: Observable) -> Observable:
        """Time shifts the observable sequence based on a subscription
        delay and a delay mapper function for each element.

        Examples:
            >>> obs = delay_with_selector(source)

        Args:
            subscription_delay: [Optional] Sequence indicating the
                delay for the subscription to the source.
            delay_duration_mapper: [Optional] Selector function to
                retrieve a sequence indicating the delay for each given
                element.

        Returns:
            Time-shifted observable sequence.
        """

        sub_delay, mapper = None, None

        if isinstance(subscription_delay, typing.Observable):
            mapper = delay_duration_mapper
            sub_delay = subscription_delay
        else:
            mapper = subscription_delay

        def subscribe(observer, scheduler=None):
            delays = CompositeDisposable()
            at_end = [False]

            def done():
                if (at_end[0] and delays.length == 0):
                    observer.on_completed()

            subscription = SerialDisposable()

            def start():
                def on_next(x):
                    try:
                        delay = mapper(x)
                    except Exception as error:
                        observer.on_error(error)
                        return

                    d = SingleAssignmentDisposable()
                    delays.add(d)

                    def on_next(_):
                        observer.on_next(x)
                        delays.remove(d)
                        done()

                    def on_completed():
                        observer.on_next(x)
                        delays.remove(d)
                        done()

                    d.disposable = delay.subscribe_(on_next, observer.on_error, on_completed, scheduler)

                def on_completed():
                    at_end[0] = True
                    subscription.dispose()
                    done()

                subscription.disposable = source.subscribe_(on_next, observer.on_error, on_completed, scheduler)

            if not sub_delay:
                start()
            else:
                subscription.disposable = sub_delay.subscribe_(lambda _: start(), observer.on_error, start)

            return CompositeDisposable(subscription, delays)
        return Observable(subscribe)
    return delay_with_mapper


from datetime import timedelta, datetime
from typing import Dict, Optional


class Cache:
    EXPIRATION_FIELD: str = "CACHE_EXPIRATION"

    def __init__(self, duration: Optional[timedelta] = None):
        self._store: Dict[str, dict] = {}
        self._duration: timedelta = duration or timedelta(minutes=30)

    def get(self, key: str, ignore_expiration: bool = False) -> Optional[dict]:
        if self.has(key) and (ignore_expiration or not self.expired(key)):
            return self._store[key]['value']
        else:
            return None

    def set(self, key: str, value: dict, duration: Optional[timedelta] = None) -> datetime:
        expiration: datetime = datetime.now() + (duration or self._duration)

        self._store[key] = {**{'value': value}, **{f'{Cache.EXPIRATION_FIELD}': expiration}}
        return expiration

    def has(self, key: str) -> bool:
        return key in self._store.keys()

    def expired(self, key: str) -> bool:
        return self.has(key) and self._store[key][Cache.EXPIRATION_FIELD] < datetime.now()

    def set_expiration(self, key: str, expiration: datetime) -> None:
        if self.has(key):
            self._store[key][Cache.EXPIRATION_FIELD] = expiration

    def extend(self, key: str, duration: timedelta) -> Optional[datetime]:
        if self.has(key):
            self._store[key][Cache.EXPIRATION_FIELD] += duration
            return self._store[key][Cache.EXPIRATION_FIELD]

    def flush(self) -> None:
        self._store.clear()

    def remove(self, key) -> None:
        if self.has(key):
            del self._store[key]

    def clean(self) -> None:
        [self.remove(key) for key in list(self._store.keys()) if self.expired(key)]



from typing import Callable, Any

from rx.core.typing import Disposable
from rx.core import Observable, typing
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable, SerialDisposable
from rx.scheduler import TimeoutScheduler


def _debounce(duetime: typing.RelativeTime, scheduler=typing.Scheduler) -> Callable[[Observable], Observable]:
    def debounce(source: Observable) -> Observable:
        """Ignores values from an observable sequence which are followed by
        another value before duetime.

        Example:
            >>> res = debounce(source)

        Args:
            source: Source observable to debounce.

        Returns:
            An operator function that takes the source observable and
            returns the debounced observable sequence.
        """

        def subscribe(observer, scheduler_=None) -> Disposable:
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()
            cancelable = SerialDisposable()
            has_value = [False]
            value = [None]
            _id = [0]

            def on_next(x: Any) -> None:
                has_value[0] = True
                value[0] = x
                _id[0] += 1
                current_id = _id[0]
                d = SingleAssignmentDisposable()
                cancelable.disposable = d

                def action(scheduler, state=None) -> None:
                    if has_value[0] and _id[0] == current_id:
                        observer.on_next(value[0])
                    has_value[0] = False

                d.disposable = _scheduler.schedule_relative(duetime, action)

            def on_error(exception: Exception) -> None:
                cancelable.dispose()
                observer.on_error(exception)
                has_value[0] = False
                _id[0] += 1

            def on_completed() -> None:
                cancelable.dispose()
                if has_value[0]:
                    observer.on_next(value[0])

                observer.on_completed()
                has_value[0] = False
                _id[0] += 1

            subscription = source.subscribe_(on_next, on_error, on_completed, scheduler=scheduler_)
            return CompositeDisposable(subscription, cancelable)
        return Observable(subscribe)
    return debounce


def _throttle_with_mapper(throttle_duration_mapper: Callable[[Any], Observable]) -> Callable[[Observable], Observable]:
    def throttle_with_mapper(source: Observable) -> Observable:
        """Partially applied throttle_with_mapper operator.

        Ignores values from an observable sequence which are followed by
        another value within a computed throttle duration.

        Example:
            >>> obs = throttle_with_mapper(source)

        Args:
            source: The observable source to throttle.

        Returns:
            The throttled observable sequence.
        """
        def subscribe(observer, scheduler=None) -> Disposable:
            cancelable = SerialDisposable()
            has_value = [False]
            value = [None]
            _id = [0]

            def on_next(x):
                throttle = None
                try:
                    throttle = throttle_duration_mapper(x)
                except Exception as e:  # pylint: disable=broad-except
                    observer.on_error(e)
                    return

                has_value[0] = True
                value[0] = x
                _id[0] += 1
                current_id = _id[0]
                d = SingleAssignmentDisposable()
                cancelable.disposable = d

                def on_next(x: Any) -> None:
                    if has_value[0] and _id[0] == current_id:
                        observer.on_next(value[0])

                    has_value[0] = False
                    d.dispose()

                def on_completed() -> None:
                    if has_value[0] and _id[0] == current_id:
                        observer.on_next(value[0])

                    has_value[0] = False
                    d.dispose()

                d.disposable = throttle.subscribe_(on_next, observer.on_error, on_completed, scheduler=scheduler)

            def on_error(e) -> None:
                cancelable.dispose()
                observer.on_error(e)
                has_value[0] = False
                _id[0] += 1

            def on_completed() -> None:
                cancelable.dispose()
                if has_value[0]:
                    observer.on_next(value[0])

                observer.on_completed()
                has_value[0] = False
                _id[0] += 1

            subscription = source.subscribe_(on_next, on_error, on_completed, scheduler=scheduler)
            return CompositeDisposable(subscription, cancelable)
        return Observable(subscribe)
    return throttle_with_mapper



class Parameters:
    def __init__(self, args: dict()):
        self.videoID : int = args['videoID']
        self.emotePopularityProcess : bool = args['emotePopularity']
        self.emoteName : str = args['emoteName']
        self.isExternal : bool = args['emoteTimeExternal']
        self.timeSegments : int = args['timeSegments'] or None
        self.numberOfTimeStamps : int = (args['timestamps'])
    def isValid(self) -> None:
        return (
            self.emotePopularityProcess 
            or self.emoteName
            )
    def commonSearchReq(self) -> None:
        return (
            (self.emoteName and not self.isExternal) 
            or self.emotePopularityProcess
            )
        



from typing import Dict, Any

import twitch.helix as helix
from twitch.api import API
from .model import Model


class Game(Model):

    def __init__(self, api: API, data: Dict[str, Any]):
        super().__init__(api, data)

        self.id: str = data.get('id')
        self.name: str = data.get('name')
        self.box_art_url: str = data.get('box_art_url')

    def __str__(self):
        return self.name

    def videos(self, **kwargs) -> 'helix.Videos':
        return helix.Videos(self._api, game_id=self.id, **kwargs)



from abc import ABCMeta
from dataclasses import dataclass
from typing import Optional, Dict, Any

from twitch.api import API


@dataclass
class Model(metaclass=ABCMeta):
    _api: Optional[API]
    data: Optional[Dict[str, Any]]



from asyncio import Future
from typing import cast, Any, Union

from rx import from_future
from rx.core import Observable, typing
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable
from rx.internal.utils import is_future


def _amb(right_source: Observable):

    if is_future(right_source):
        obs = from_future(cast(Future, right_source))
    else:
        obs = cast(Observable, right_source)

    def amb(left_source: Observable):
        def subscribe(observer: typing.Observer, scheduler: typing.Scheduler = None) -> typing.Disposable:
            choice = [None]
            left_choice = 'L'
            right_choice = 'R'
            left_subscription = SingleAssignmentDisposable()
            right_subscription = SingleAssignmentDisposable()

            def choice_left():
                if not choice[0]:
                    choice[0] = left_choice
                    right_subscription.dispose()

            def choice_right():
                if not choice[0]:
                    choice[0] = right_choice
                    left_subscription.dispose()

            def on_next_left(value):
                with left_source.lock:
                    choice_left()
                if choice[0] == left_choice:
                    observer.on_next(value)

            def on_error_left(err):
                with left_source.lock:
                    choice_left()
                if choice[0] == left_choice:
                    observer.on_error(err)

            def on_completed_left():
                with left_source.lock:
                    choice_left()
                if choice[0] == left_choice:
                    observer.on_completed()

            left_d = left_source.subscribe_(on_next_left, on_error_left, on_completed_left, scheduler)
            left_subscription.disposable = left_d

            def send_right(value: Any) -> None:
                with left_source.lock:
                    choice_right()
                if choice[0] == right_choice:
                    observer.on_next(value)

            def on_error_right(err: Exception) -> None:
                with left_source.lock:
                    choice_right()
                if choice[0] == right_choice:
                    observer.on_error(err)

            def on_completed_right() -> None:
                with left_source.lock:
                    choice_right()
                if choice[0] == right_choice:
                    observer.on_completed()

            right_d = obs.subscribe_(send_right, on_error_right, on_completed_right, scheduler)
            right_subscription.disposable = right_d
            return CompositeDisposable(left_subscription, right_subscription)
        return Observable(subscribe)
    return amb


from asyncio import Future
from typing import cast, Callable, Union

from rx import from_future
from rx.internal import noop
from rx.core import Observable
from rx.disposable import CompositeDisposable
from rx.internal.utils import is_future


def _take_until(other: Union[Observable, Future]) -> Callable[[Observable], Observable]:
    if is_future(other):
        obs = from_future(cast(Future, other))
    else:
        obs = cast(Observable, other)

    def take_until(source: Observable) -> Observable:
        """Returns the values from the source observable sequence until
        the other observable sequence produces a value.

        Args:
            source: The source observable sequence.

        Returns:
            An observable sequence containing the elements of the source
            sequence up to the point the other sequence interrupted
            further propagation.
        """

        def subscribe(observer, scheduler=None):

            def on_completed(_):
                observer.on_completed()

            return CompositeDisposable(
                source.subscribe(observer),
                obs.subscribe_(on_completed, observer.on_error, noop, scheduler)
            )
        return Observable(subscribe)
    return take_until



from matplotlib import _api
from ._formsubplottool import UiSubplotTool


_api.warn_deprecated(
    "3.3", obj_type="module", name=__name__,
    alternative="matplotlib.backends.backend_qt5.SubplotToolQt")
__all__ = ["UiSubplotTool"]



from typing import Callable, Optional
from datetime import datetime, timedelta

from rx import operators as ops
from rx.core import Observable, typing
from rx.internal.constants import DELTA_ZERO
from rx.disposable import CompositeDisposable, SerialDisposable, MultipleAssignmentDisposable
from rx.scheduler import TimeoutScheduler


class Timestamp(object):
    def __init__(self, value, timestamp):
        self.value = value
        self.timestamp = timestamp


def observable_delay_timespan(source: Observable, duetime: typing.RelativeTime,
                              scheduler: Optional[typing.Scheduler] = None) -> Observable:

    def subscribe(observer, scheduler_=None):
        nonlocal duetime

        _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()

        if isinstance(duetime, datetime):
            duetime = _scheduler.to_datetime(duetime) - _scheduler.now
        else:
            duetime = _scheduler.to_timedelta(duetime)

        cancelable = SerialDisposable()
        exception = [None]
        active = [False]
        running = [False]
        queue = []

        def on_next(notification):
            should_run = False

            with source.lock:
                if notification.value.kind == 'E':
                    del queue[:]
                    queue.append(notification)
                    exception[0] = notification.value.exception
                    should_run = not running[0]
                else:
                    queue.append(Timestamp(value=notification.value, timestamp=notification.timestamp + duetime))
                    should_run = not active[0]
                    active[0] = True

            if should_run:
                if exception[0]:
                    observer.on_error(exception[0])
                else:
                    mad = MultipleAssignmentDisposable()
                    cancelable.disposable = mad

                    def action(scheduler, state):
                        if exception[0]:
                            return

                        with source.lock:
                            running[0] = True
                            while True:
                                result = None
                                if queue and queue[0].timestamp <= scheduler.now:
                                    result = queue.pop(0).value

                                if result:
                                    result.accept(observer)

                                if not result:
                                    break

                            should_continue = False
                            recurse_duetime = 0
                            if queue:
                                should_continue = True
                                diff = queue[0].timestamp - scheduler.now
                                zero = DELTA_ZERO if isinstance(diff, timedelta) else 0
                                recurse_duetime = max(zero, diff)
                            else:
                                active[0] = False

                            ex = exception[0]
                            running[0] = False

                        if ex:
                            observer.on_error(ex)
                        elif should_continue:
                            mad.disposable = scheduler.schedule_relative(recurse_duetime, action)

                    mad.disposable = _scheduler.schedule_relative(duetime, action)
        subscription = source.pipe(
            ops.materialize(),
            ops.timestamp()
        ).subscribe_(on_next, scheduler=_scheduler)

        return CompositeDisposable(subscription, cancelable)
    return Observable(subscribe)


def _delay(duetime: typing.RelativeTime, scheduler: Optional[typing.Scheduler] = None) -> Callable[[Observable], Observable]:
    def delay(source: Observable) -> Observable:
        """Time shifts the observable sequence.

        A partially applied delay operator function.

        Examples:
            >>> res = delay(source)

        Args:
            source: The observable sequence to delay.

        Returns:
            A time-shifted observable sequence.
        """
        return observable_delay_timespan(source, duetime, scheduler)
    return delay



from threading import RLock
from typing import Optional, List

from rx import from_future
from rx.core import Observable, typing
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable
from rx.internal.concurrency import synchronized
from rx.internal.utils import is_future


# pylint: disable=redefined-builtin

def _zip(*args: Observable) -> Observable:
    """Merges the specified observable sequences into one observable
    sequence by creating a tuple whenever all of the
    observable sequences have produced an element at a corresponding
    index.

    Example:
        >>> res = zip(obs1, obs2)

    Args:
        args: Observable sources to zip.

    Returns:
        An observable sequence containing the result of combining
        elements of the sources as tuple.
    """

    sources = list(args)

    def subscribe(observer: typing.Observer,
                  scheduler: Optional[typing.Scheduler] = None) -> CompositeDisposable:
        n = len(sources)
        queues: List[List] = [[] for _ in range(n)]
        lock = RLock()

        @synchronized(lock)
        def next(i):
            if all([len(q) for q in queues]):
                try:
                    queued_values = [x.pop(0) for x in queues]
                    res = tuple(queued_values)
                except Exception as ex:  # pylint: disable=broad-except
                    observer.on_error(ex)
                    return

                observer.on_next(res)

        subscriptions = [None] * n

        def func(i):
            source = sources[i]
            sad = SingleAssignmentDisposable()
            source = from_future(source) if is_future(source) else source

            def on_next(x):
                queues[i].append(x)
                next(i)

            sad.disposable = source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
            subscriptions[i] = sad

        for idx in range(n):
            func(idx)
        return CompositeDisposable(subscriptions)

    return Observable(subscribe)



import threading
from typing import Any, List, Optional

from rx.disposable import Disposable
from rx.core import Observable, Observer, typing
from rx.internal import DisposedException

from .innersubscription import InnerSubscription


class Subject(Observable, Observer, typing.Subject):
    """Represents an object that is both an observable sequence as well
    as an observer. Each notification is broadcasted to all subscribed
    observers.
    """

    def __init__(self) -> None:
        super().__init__()

        self.is_disposed = False
        self.observers: List[typing.Observer] = []
        self.exception: Optional[Exception] = None

        self.lock = threading.RLock()

    def check_disposed(self) -> None:
        if self.is_disposed:
            raise DisposedException()

    def _subscribe_core(self,
                        observer: typing.Observer,
                        scheduler: Optional[typing.Scheduler] = None
                        ) -> typing.Disposable:
        with self.lock:
            self.check_disposed()
            if not self.is_stopped:
                self.observers.append(observer)
                return InnerSubscription(self, observer)

            if self.exception is not None:
                observer.on_error(self.exception)
            else:
                observer.on_completed()
            return Disposable()

    def on_next(self, value: Any) -> None:
        """Notifies all subscribed observers with the value.

        Args:
            value: The value to send to all subscribed observers.
        """

        with self.lock:
            self.check_disposed()
        super().on_next(value)

    def _on_next_core(self, value: Any) -> None:
        with self.lock:
            observers = self.observers.copy()

        for observer in observers:
            observer.on_next(value)

    def on_error(self, error: Exception) -> None:
        """Notifies all subscribed observers with the exception.

        Args:
            error: The exception to send to all subscribed observers.
        """

        with self.lock:
            self.check_disposed()
        super().on_error(error)

    def _on_error_core(self, error: Exception) -> None:
        with self.lock:
            observers = self.observers.copy()
            self.observers.clear()
            self.exception = error

        for observer in observers:
            observer.on_error(error)

    def on_completed(self) -> None:
        """Notifies all subscribed observers of the end of the sequence."""

        with self.lock:
            self.check_disposed()
        super().on_completed()

    def _on_completed_core(self) -> None:
        with self.lock:
            observers = self.observers.copy()
            self.observers.clear()

        for observer in observers:
            observer.on_completed()

    def dispose(self) -> None:
        """Unsubscribe all observers and release resources."""

        with self.lock:
            self.is_disposed = True
            self.observers = []
            self.exception = None
            super().dispose()



import tracemalloc
import unittest
from typing import Dict, Any

import responses

import twitch

tracemalloc.start()


class TestHelixUser(unittest.TestCase):
    user: Dict[str, Dict[str, Any]] = {
        'zarlach': {
            'id': '24250859',
            'login': 'zarlach',
            'display_name': 'Zarlach',
            'type': '',
            'broadcaster_type': '',
            'description': '',
            'profile_image_url': 'https://static-cdn.jtvnw.net/jtv_user_pictures/zarlach-profile_image-1cb98e7eadb5918a-300x300.png',
            'offline_image_url': 'https://static-cdn.jtvnw.net/jtv_user_pictures/zarlach-channel_offline_image-f2d036ac9582d793-1920x1080.png',
            'view_count': 1664
        },
        'sodapoppin': {
            'id': '26301881',
            'login': 'sodapoppin',
            'display_name': 'sodapoppin',
            'type': '',
            'broadcaster_type': 'partner',
            'description': 'Wtf do i write here? Click my stream, or i scream.',
            'profile_image_url': 'https://static-cdn.jtvnw.net/jtv_user_pictures/sodapoppin-profile_image-10049b6200f90c14-300x300.png',
            'offline_image_url': 'https://static-cdn.jtvnw.net/jtv_user_pictures/7ed72b04-897e-4a85-a3c9-2a8ba74aeaa7-channel_offline_image-1920x1080.jpg',
            'view_count': 276164750
        }
    }

    def setUp(self) -> None:
        responses.add(responses.GET, 'https://id.twitch.tv/oauth2/token?client_token=id&client_secret=secret&grant_type=client_credentials',
                      match_querystring=True,
                      json={'access_token': 'token'})

        responses.add(responses.GET, 'https://api.twitch.tv/helix/users?login=zarlach',
                      match_querystring=True,
                      json={'data': [TestHelixUser.user['zarlach']]})
        responses.add(responses.GET, 'https://api.twitch.tv/helix/users?login=sodapoppin',
                      match_querystring=True,
                      json={'data': [TestHelixUser.user['sodapoppin']]})

        responses.add(responses.GET, 'https://api.twitch.tv/helix/users?id=24250859&login=sodapoppin',
                      match_querystring=True,
                      json={'data': [TestHelixUser.user['zarlach'], TestHelixUser.user['sodapoppin']]})

        responses.add(responses.GET, 'https://api.twitch.tv/helix/users?login=sodapoppin&login=zarlach',
                      match_querystring=True,
                      json={'data': [TestHelixUser.user['zarlach'], TestHelixUser.user['sodapoppin']]})

    @responses.activate
    def test_user(self):
        helix = twitch.Helix(client_id='id', bearer_token='token', use_cache=True)

        # Get display name to display name
        self.assertEqual(helix.user('zarlach').display_name, 'Zarlach')

    @responses.activate
    def test_users(self):
        # Should returned cached data from above
        helix = twitch.Helix(client_id='id', bearer_token='token', use_cache=True)

        for user, display_name in zip(helix.users([24250859, 'sodapoppin']), ['Zarlach', 'sodapoppin']):
            self.assertEqual(user.display_name, display_name)

    @responses.activate
    def test_custom_user_cache(self):
        helix = twitch.Helix(client_id='id', bearer_token='token', use_cache=True)
        helix.users(['zarlach', 'sodapoppin'])

        # Users have custom caching, such that url should not be cached
        self.assertFalse(helix.api.SHARED_CACHE.has('GET:https://api.twitch.tv/helix/users?login=zarlach'))

        # Cache entries by login name and id numberit
        self.assertTrue(helix.api.SHARED_CACHE.has('helix.users.login.zarlach'))
        self.assertTrue(helix.api.SHARED_CACHE.has('helix.users.id.24250859'))

        self.assertTrue(helix.api.SHARED_CACHE.has('helix.users.login.sodapoppin'))
        self.assertTrue(helix.api.SHARED_CACHE.has('helix.users.id.26301881'))

        # Flush cache to remove
        helix.api.flush_cache()

        # Check cache flush
        self.assertFalse(helix.api.SHARED_CACHE.has('helix.users.login.zarlach'))
        self.assertFalse(helix.api.SHARED_CACHE.has('helix.users.id.24250859'))

        self.assertFalse(helix.api.SHARED_CACHE.has('helix.users.login.sodapoppin'))
        self.assertFalse(helix.api.SHARED_CACHE.has('helix.users.id.26301881'))



from typing import Dict, Any

import twitch.helix as helix
import twitch.tmi as tmi
from twitch.api import API
from .model import Model


class User(Model):

    def __init__(self, api: API, data: Dict[str, Any]):
        super().__init__(api, data)

        self.broadcaster_type: str = data.get('broadcaster_type')
        self.description: str = data.get('description')
        self.display_name: str = data.get('display_name')
        self.email: str = data.get('email')
        self.id: str = data.get('id')
        self.login: str = data.get('login')
        self.offline_image_url: str = data.get('offline_image_url')
        self.profile_image_url: str = data.get('profile_image_url')
        self.type: str = data.get('type')
        self.view_count: int = data.get('view_count')

    def __str__(self):
        return self.login

    def videos(self, **kwargs) -> 'helix.Videos':
        return helix.Videos(api=self._api, user_id=int(self.id), **kwargs)

    @property
    def stream(self) -> 'helix.Stream':
        return helix.Streams(api=self._api, user_id=int(self.id))[0]

    @property
    def is_live(self) -> bool:
        try:
            if self.stream:
                return True
        except helix.StreamNotFound:
            return False

    @property
    def chatters(self) -> 'tmi.Chatters':
        source = tmi.TMI('')
        source.api = self._api
        source.api.base_url = tmi.TMI.BASE_URL

        return source.chatters(self.login)

    def following(self, **kwargs) -> 'helix.Follows':
        kwargs['from_id'] = self.id
        return helix.Follows(api=self._api, follow_type='followings', **kwargs)

    def followers(self, **kwargs) -> 'helix.Follows':
        kwargs['to_id'] = self.id
        return helix.Follows(api=self._api, follow_type='followers', **kwargs)



from typing import Callable, Optional

from rx.core import Observable, typing
from rx.scheduler import TimeoutScheduler
from rx.internal.utils import add_ref
from rx.disposable import SingleAssignmentDisposable, CompositeDisposable, RefCountDisposable, SerialDisposable
from rx.subject import Subject


def _window_with_time_or_count(timespan: typing.RelativeTime, count: int, scheduler: Optional[typing.Scheduler] = None
                               ) -> Callable[[Observable], Observable]:
    def window_with_time_or_count(source: Observable) -> Observable:
        def subscribe(observer, scheduler_=None):
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()

            n = [0]
            s = [None]
            timer_d = SerialDisposable()
            window_id = [0]
            group_disposable = CompositeDisposable(timer_d)
            ref_count_disposable = RefCountDisposable(group_disposable)

            def create_timer(_id):
                m = SingleAssignmentDisposable()
                timer_d.disposable = m

                def action(scheduler, state):
                    if _id != window_id[0]:
                        return

                    n[0] = 0
                    window_id[0] += 1
                    new_id = window_id[0]
                    s[0].on_completed()
                    s[0] = Subject()
                    observer.on_next(add_ref(s[0], ref_count_disposable))
                    create_timer(new_id)

                m.disposable = _scheduler.schedule_relative(timespan, action)

            s[0] = Subject()
            observer.on_next(add_ref(s[0], ref_count_disposable))
            create_timer(0)

            def on_next(x):
                new_window = False
                new_id = 0

                s[0].on_next(x)
                n[0] += 1
                if n[0] == count:
                    new_window = True
                    n[0] = 0
                    window_id[0] += 1
                    new_id = window_id[0]
                    s[0].on_completed()
                    s[0] = Subject()
                    observer.on_next(add_ref(s[0], ref_count_disposable))

                if new_window:
                    create_timer(new_id)

            def on_error(e):
                s[0].on_error(e)
                observer.on_error(e)

            def on_completed():
                s[0].on_completed()
                observer.on_completed()

            group_disposable.add(source.subscribe_(on_next, on_error, on_completed, scheduler_))
            return ref_count_disposable
        return Observable(subscribe)
    return window_with_time_or_count



from typing import Callable
from rx.core import Observable


def _to_iterable() -> Callable[[Observable], Observable]:
    def to_iterable(source: Observable) -> Observable:
        """Creates an iterable from an observable sequence.

        Returns:
            An observable sequence containing a single element with an
            iterable containing all the elements of the source
            sequence.
        """
        def subscribe(observer, scheduler=None):
            nonlocal source

            queue = []

            def on_next(item):
                queue.append(item)

            def on_completed():
                nonlocal queue
                observer.on_next(queue)
                queue = []
                observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return to_iterable


from typing import Callable
from asyncio import Future

from rx import throw, from_future
from rx.core import Observable


def _start_async(function_async: Callable[[], Future]) -> Observable:
    try:
        future = function_async()
    except Exception as ex:  # pylint: disable=broad-except
        return throw(ex)

    return from_future(future)



from typing import List, Callable

from twitch.tmi.models.chatter import Chatter
from twitch.tmi.resources.chatters import Chatters
from twitch.tmi.tmi import TMI

__all__: List[Callable] = [
    TMI,
    Chatter, Chatters
]



from typing import Callable, Any, TypeVar, overload
from functools import reduce

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
E = TypeVar('E')
F = TypeVar('F')
G = TypeVar('G')


@overload
def pipe(*operators: Callable[['Observable'], 'Observable']) -> Callable[['Observable'], 'Observable']:  # type: ignore
    """Compose multiple operators left to right.

    Composes zero or more operators into a functional composition. The
    operators are composed to left to right. A composition of zero
    operators gives back the source.

    Examples:
        >>> pipe()(source) == source
        >>> pipe(f)(source) == f(source)
        >>> pipe(f, g)(source) == g(f(source))
        >>> pipe(f, g, h)(source) == h(g(f(source)))
        ...

    Returns:
        The composed observable.
    """
    ...

@overload
def pipe() -> Callable[[A], A]:  # pylint: disable=function-redefined
    ...  # pylint: disable=pointless-statement


@overload
def pipe(op1: Callable[[A], B]) -> Callable[[A], B]:  # pylint: disable=function-redefined
    ...  # pylint: disable=pointless-statement


@overload
def pipe(op1: Callable[[A], B], op2: Callable[[B], C]) -> Callable[[A], C]:  # pylint: disable=function-redefined
    ...  # pylint: disable=pointless-statement


@overload
def pipe(op1: Callable[[A], B],  # pylint: disable=function-redefined
         op2: Callable[[B], C],
         op3: Callable[[C], D]
         ) -> Callable[[A], D]:
    ...  # pylint: disable=pointless-statement


@overload
def pipe(op1: Callable[[A], B],  # pylint: disable=function-redefined
         op2: Callable[[B], C],
         op3: Callable[[C], D],
         op4: Callable[[D], E]
         ) -> Callable[[A], E]:
    ...  # pylint: disable=pointless-statement


@overload
def pipe(op1: Callable[[A], B],  # pylint: disable=function-redefined
         op2: Callable[[B], C],
         op3: Callable[[C], D],
         op4: Callable[[D], E],
         op5: Callable[[E], F]
         ) -> Callable[[A], F]:
    ...  # pylint: disable=pointless-statement


@overload
def pipe(op1: Callable[[A], B],  # pylint: disable=function-redefined,too-many-arguments
         op2: Callable[[B], C],
         op3: Callable[[C], D],
         op4: Callable[[D], E],
         op5: Callable[[E], F],
         op6: Callable[[F], G]
         ) -> Callable[[A], G]:
    ...  # pylint: disable=pointless-statement


# pylint: disable=function-redefined
def pipe(*operators: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """Compose multiple operators left to right.

    Composes zero or more operators into a functional composition. The
    operators are composed to left to right. A composition of zero
    operators gives back the source.

    Examples:
        >>> pipe()(source) == source
        >>> pipe(f)(source) == f(source)
        >>> pipe(f, g)(source) == g(f(source))
        >>> pipe(f, g, h)(source) == h(g(f(source)))
        ...

    Returns:
        The composed observable.
    """

    def compose(source: Any) -> Any:
        return reduce(lambda obs, op: op(obs), operators, source)
    return compose



from typing import Callable

from rx.disposable import Disposable
from rx.core import Observable


def _finally_action(action: Callable) -> Callable[[Observable], Observable]:
    def finally_action(source: Observable) -> Observable:
        """Invokes a specified action after the source observable
        sequence terminates gracefully or exceptionally.

        Example:
            res = finally(source)

        Args:
            source: Observable sequence.

        Returns:
            An observable sequence with the action-invoking termination
            behavior applied.
        """

        def subscribe(observer, scheduler=None):
            try:
                subscription = source.subscribe(observer, scheduler=scheduler)
            except Exception:
                action()
                raise

            def dispose():
                try:
                    subscription.dispose()
                finally:
                    action()

            return Disposable(dispose)
        return Observable(subscribe)
    return finally_action



from typing import List, Optional, Dict, Any

import twitch.helix as helix
from twitch.api import API
from .model import Model


class Commenter:

    def __init__(self, data: Dict[str, Any]):
        self.data: Dict[str, Any] = data

        self.display_name: str = data.get('display_name')
        self.id: str = data.get('_id')
        self.name: str = data.get('name')
        self.type: str = data.get('type')
        self.bio: str = data.get('bio')
        self.created_at: str = data.get('created_at')
        self.updated_at: str = data.get('updated_at')
        self.logo: str = data.get('logo')


class Emoticon:

    def __init__(self, data: Dict[str, Any]):
        self.data: Dict[str, Any] = data

        self.id: str = data.get('_id')
        self.begin: int = data.get('begin')
        self.end: int = data.get('end')
        self.emoticon_id: str = data.get('emoticon_id')
        self.emoticon_set_id: str = data.get('emoticon_set_id')


class Fragment:

    def __init__(self, data: Optional[Dict[str, Any]] = None):
        self.data: Dict[str, Any] = data

        self.text: str = data.get('text')
        self.emoticon: Optional[Emoticon] = Emoticon(data.get('emoticon')) if data.get('emoticon') else None


class UserBadge:

    def __init__(self, data: Dict[str, Any]):
        self.data: Dict[str, Any] = data

        self.id: str = data.get('_id')
        self.version: str = data.get('version')


class Message:

    def __init__(self, data: Dict[str, Any]):
        self.data: Dict[str, Any] = data

        self.body: str = data.get('body')
        self.emoticons: List[Emoticon] = [Emoticon(data) for data in data.get('emoticons', [])]
        self.fragments: List[Fragment] = [Fragment(data) for data in data.get('fragments', [])]
        self.is_action: bool = data.get('is_action')
        self.user_badges: List[UserBadge] = [UserBadge(data) for data in data.get('user_badges', [])]
        self.user_color: str = data.get('user_color')


class Comment(Model):

    def __init__(self, api: API, data: Dict[str, Any]):
        super().__init__(api, data)

        self.id: str = data.get('_id')
        self.created_at: str = data.get('created_at')
        self.updated_at: str = data.get('updated_at')
        self.channel_id: str = data.get('channel_id')
        self.content_type: str = data.get('content_type')
        self.content_id: str = data.get('content_id')
        self.content_offset_seconds: float = data.get('content_offset_seconds')
        self.commenter: Commenter = Commenter(data.get('commenter'))
        self.source: str = data.get('source')
        self.state: str = data.get('state')
        self.message: Message = Message(data.get('message'))
        self.more_replies: bool = data.get('more_replies')

    @property
    def user(self) -> 'helix.User':
        return helix.Helix(client_id=self._api.client_id).user(int(self.commenter.id))



from typing import Any, Callable

from rx import operators as ops
from rx.core import Observable


def _pluck(key: Any) -> Callable[[Observable], Observable]:
    """Retrieves the value of a specified key using dict-like access (as in
    element[key]) from all elements in the Observable sequence.

    Args:
        key: The key to pluck.

    Returns a new Observable {Observable} sequence of key values.

    To pluck an attribute of each element, use pluck_attr.
    """

    return ops.map(lambda x: x[key])


def _pluck_attr(prop: str) -> Callable[[Observable], Observable]:
    """Retrieves the value of a specified property (using getattr) from
    all elements in the Observable sequence.

    Args:
        property: The property to pluck.

    Returns a new Observable {Observable} sequence of property values.

    To pluck values using dict-like access (as in element[key]) on each
    element, use pluck.
    """

    return ops.map(lambda x: getattr(x, prop))



from typing import Callable, Optional, Any

import rx
from rx.core import Observable
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable, SerialDisposable


def _timeout_with_mapper(first_timeout: Optional[Observable] = None,
                         timeout_duration_mapper: Optional[Callable[[Any], Observable]] = None,
                         other: Optional[Observable] = None
                         ) -> Callable[[Observable], Observable]:
    """Returns the source observable sequence, switching to the other
    observable sequence if a timeout is signaled.

        res = timeout_with_mapper(rx.timer(500))
        res = timeout_with_mapper(rx.timer(500), lambda x: rx.timer(200))
        res = timeout_with_mapper(rx.timer(500), lambda x: rx.timer(200)), rx.return_value(42))

    Args:
        first_timeout -- [Optional] Observable sequence that represents the
            timeout for the first element. If not provided, this defaults to
            rx.never().
        timeout_duration_mapper -- [Optional] Selector to retrieve an
            observable sequence that represents the timeout between the
            current element and the next element.
        other -- [Optional] Sequence to return in case of a timeout. If not
            provided, this is set to rx.throw().

    Returns:
        The source sequence switching to the other sequence in case
    of a timeout.
    """

    first_timeout = first_timeout or rx.never()
    other = other or rx.throw(Exception('Timeout'))

    def timeout_with_mapper(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            subscription = SerialDisposable()
            timer = SerialDisposable()
            original = SingleAssignmentDisposable()

            subscription.disposable = original

            switched = False
            _id = [0]

            def set_timer(timeout: Observable) -> None:
                my_id = _id[0]

                def timer_wins():
                    return _id[0] == my_id

                d = SingleAssignmentDisposable()
                timer.disposable = d

                def on_next(x):
                    if timer_wins():
                        subscription.disposable = other.subscribe(observer, scheduler=scheduler)

                    d.dispose()

                def on_error(e):
                    if timer_wins():
                        observer.on_error(e)

                def on_completed():
                    if timer_wins():
                        subscription.disposable = other.subscribe(observer)

                d.disposable = timeout.subscribe_(on_next, on_error, on_completed, scheduler)

            set_timer(first_timeout)

            def observer_wins():
                res = not switched
                if res:
                    _id[0] += 1

                return res

            def on_next(x):
                if observer_wins():
                    observer.on_next(x)
                    timeout = None
                    try:
                        timeout = timeout_duration_mapper(x)
                    except Exception as e:
                        observer.on_error(e)
                        return

                    set_timer(timeout)

            def on_error(error):
                if observer_wins():
                    observer.on_error(error)

            def on_completed():
                if observer_wins():
                    observer.on_completed()

            original.disposable = source.subscribe_(on_next, on_error, on_completed, scheduler)
            return CompositeDisposable(subscription, timer)
        return Observable(subscribe)
    return timeout_with_mapper



from typing import Optional

from rx.disposable import Disposable
from rx.core import Observable, typing


def _never() -> Observable:
    """Returns a non-terminating observable sequence, which can be used
    to denote an infinite duration (e.g. when using reactive joins).

    Returns:
        An observable sequence whose observers will never get called.
    """

    def subscribe(observer: typing.Observer, scheduler: Optional[typing.Scheduler] = None) -> typing.Disposable:
        return Disposable()

    return Observable(subscribe)



from typing import Callable
from rx.core import Observable


def _take_last_buffer(count: int) -> Callable[[Observable], Observable]:
    def take_last_buffer(source: Observable) -> Observable:
        """Returns an array with the specified number of contiguous
        elements from the end of an observable sequence.

        Example:
            >>> res = take_last(source)

        This operator accumulates a buffer with a length enough to
        store elements count elements. Upon completion of the source
        sequence, this buffer is drained on the result sequence. This
        causes the elements to be delayed.

        Args:
            source: Source observable to take elements from.

        Returns:
            An observable sequence containing a single list with the
            specified number of elements from the end of the source
            sequence.
        """

        def subscribe(observer, scheduler=None):
            q = []

            def on_next(x):
                with source.lock:
                    q.append(x)
                    if len(q) > count:
                        q.pop(0)

            def on_completed():
                observer.on_next(q)
                observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return take_last_buffer



from typing import List, Dict, Any

import twitch.helix as helix
from twitch.api import API
from .model import Model


class Stream(Model):

    def __init__(self, api: API, data: Dict[str, Any]):
        super().__init__(api, data)

        self.id: str = data.get('id')
        self.user_id: str = data.get('user_id')
        self.game_id: str = data.get('game_id')
        self.community_ids: List[str] = data.get('community_ids', [])
        self.type: str = data.get('type')
        self.title: str = data.get('title')
        self.viewer_count: int = data.get('viewer_count')
        self.started_at: str = data.get('started_at')
        self.language: str = data.get('language')
        self.thumbnail_url: str = data.get('thumbnail_url')

    def __str__(self):
        return self.title

    @property
    def user(self) -> 'helix.User':
        return helix.Users(self._api, int(self.user_id))[0]
